#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################

"""
Basic IDAES unit model for compressed hydrogen gas (chg) tank.
Uses custom material and energy balance equations.
Suitable for use with steady state flowsheets indexed with time.

A previous state block (previous_state) is defined
to pass the previous time point or starting condition of the tank,
 i.e., Pressure and Temperature.
For dynamic operations, multiple instances of this model indexed with time
can be used. For this, equality constraints must be written at
flowsheet level, i.e., 
the final state of tank (Pres & Temp) at time (t-1) ==
    intial state of tank (Pres & Temp) at time (t)

This property package assumes Ideal gas behaviour of hydrogen gas.
Adiabatic operations are assumed.
TODO: add constraints to enable isothermal operations
TODO: add capability to allow liquid phase storage
"""
from collections import OrderedDict
from pandas import DataFrame
import textwrap
import sys
# Import Pyomo libraries
from pyomo.environ import (Var,
                           Reals,
                           NonNegativeReals,
                           Constraint,
                           value
                           )
# from pyomo.network import Port
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES modules
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialFlowBasis,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.constants import Constants as const
from idaes.core.solvers import get_solver
from idaes.core.util.tables import stream_table_dataframe_to_string

import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale

__author__ = "Naresh Susarla"

# Set up logger
_log = idaeslog.getLogger(__name__)

@declare_process_block_class("HydrogenTank",
                             doc="Simple compressed hydrogen tank model")
class HydrogenTankData(UnitModelBlockData):
    """
    Simple hydrogen tank model.
    Unit model to store or supply compressed hydrogen.
    
    """
    CONFIG = ConfigBlock()
    
    # This model is based on steady state material & energy balances.
    # The accumulation term is computed based on the tank state at
    # previous time step. Thus, dynamic option is turned off.
    # However, a dynamic analysis can be performed by creating
    # an instance of this model for every time step.
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicats if Hydrogen tank model is dynamic,
**default** = False. Equilibrium Reactors do not support dynamic behavior."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Indicates whether holdup terms should be constructed or not.
**default** - False. Hydrogen tank model uses custom equations for holdup."""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))

    def build(self):
        """Building model
        Args:
            None
        Returns:
            None
        """
        super().build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        # add inlet and outlet states
        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        # add tank volume
        self.control_volume.add_geometry()

        # add phase fractions
        self.control_volume._add_phase_fractions()

        # add pressure balance
        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=True)

        # add a state block 'previous_state' for the storage tank
        # this state block is needed to compute material and energy holdup
        # at previous or the starting time step using the property package
        # for a given P_prev and T_prev
        # NOTE: there is no flow in the previous state so,
        # flow_mol state variable is fixed to 0
        self.previous_state = (
            self.config.property_package.build_state_block(
                self.flowsheet().config.time,
                doc="Tank state at previous time"))

        # previous state should not have any flow
        self.previous_state[:].flow_mol.fix(0)

        # add local lists for easy use
        phase_list = self.control_volume.properties_in.phase_list
        pc_set = self.control_volume.properties_in.phase_component_set
        component_list = self.control_volume.properties_in.component_list

        # Get units from property package
        units = self.config.property_package.\
            get_metadata().get_derived_units
        if (self.control_volume.properties_in[
                self.flowsheet().config.time.first()]
                .get_material_flow_basis() == MaterialFlowBasis.molar):
            flow_units = units("flow_mole")
            material_units = units("amount")
        elif (self.control_volume.properties_in[
                self.flowsheet().config.time.first()]
              .get_material_flow_basis() == MaterialFlowBasis.mass):
            flow_units = units("flow_mass")
            material_units = units("mass")

        # Add Inlet and Outlet Ports
        self.add_inlet_port()
        self.add_outlet_port()

        # Define Vars for Tank volume calculations
        self.tank_diameter = Var(
            self.flowsheet().config.time,
            within=NonNegativeReals,
            initialize=1.0,
            doc="Diameter of storage tank in m",
            units=units("length"))

        self.tank_length = Var(
            self.flowsheet().config.time,
            within=NonNegativeReals,
            initialize=1.0,
            doc="Length of storage tank in m",
            units=units("length"))

        # Tank volume calculation
        @self.Constraint(self.flowsheet().config.time)
        def volume_cons(b, t):
            return (b.control_volume.volume[t] ==
                    const.pi * b.tank_length[t] *
                    ((b.tank_diameter[t]/2)**2))

        # define Vars for the model
        self.dt = Var(
            self.flowsheet().config.time,
            domain=NonNegativeReals,
            initialize=100,
            doc="Time step for holdup calculation",
            units=units("time"))

        self.heat_duty = Var(
            self.flowsheet().config.time,
            domain=Reals,
            initialize=0.0,
            doc="Heat transferred from surroundings, 0 for adiabatic",
            units=units("power"))

        self.material_accumulation = Var(
            self.flowsheet().config.time,
            pc_set,
            within=Reals,
            initialize=1.0,
            doc="Accumulation of material in tank",
            units=flow_units)

        self.energy_accumulation = Var(
            self.flowsheet().config.time,
            phase_list,
            within=Reals,
            initialize=1.0,
            doc="Energy accumulation",
            units=units("power"))

        self.material_holdup = Var(
            self.flowsheet().config.time,
            pc_set,
            within=NonNegativeReals,
            initialize=1.0,
            doc="Material holdup in tank",
            units=material_units)

        self.energy_holdup = Var(
            self.flowsheet().config.time,
            phase_list,
            within=NonNegativeReals,
            initialize=1.0,
            doc="Energy holdup in tank",
            units=units("energy"))

        self.previous_material_holdup = Var(
            self.flowsheet().config.time,
            pc_set,
            within=NonNegativeReals,
            initialize=1.0,
            doc="Tank material holdup at previous time",
            units=material_units)

        self.previous_energy_holdup = Var(
            self.flowsheet().config.time,
            phase_list,
            within=NonNegativeReals,
            initialize=1.0,
            doc="Tank energy holdup at previous time",
            units=units("energy"))

        # Adiabatic operations are assumed
        # Fixing the heat_duty to 0 here to avoid any misakes at use
        # TODO: remove this once the isothermal constraints are added
        self.heat_duty.fix(0)

        # Computing material and energy holdup in the tank at previous time
        # using previous state Pressure and Temperature of the tank
        @self.Constraint(self.flowsheet().config.time,
                         pc_set,
                         doc="Material holdup at previous time")
        def previous_material_holdup_rule(b, t, p, j):
            return (
                b.previous_material_holdup[t, p, j]
                == b.control_volume.volume[t] *
                b.control_volume.phase_fraction[t, p] *
                b.previous_state[t].get_material_density_terms(p, j)
                )

        @self.Constraint(self.flowsheet().config.time,
                         phase_list,
                         doc="Energy holdup at previous time")
        def previous_energy_holdup_rule(b, t, p):
            if (self.control_volume.properties_in[t]
                .get_material_flow_basis() == MaterialFlowBasis.molar):
                return (
                    b.previous_energy_holdup[t, p] ==
                    (sum(b.previous_material_holdup[t, p, j]
                         for j in component_list)
                    * b.previous_state[t].energy_internal_mol_phase[p])
                    )
            if (self.control_volume.properties_in[t]
                .get_material_flow_basis() == MaterialFlowBasis.mass):
                return (
                    b.previous_energy_holdup[t, p] ==
                    (sum(b.previous_material_holdup[t, p, j]
                         for j in component_list)
                    * (b.previous_state[t].energy_internal_mol_phase[p]/
                       b.previous_state[t].mw))
                    )

        # component material balances
        @self.Constraint(self.flowsheet().config.time,
                         pc_set,
                         doc="Material balances")
        def material_balances(b, t, p, j):
            if (p, j) in pc_set:
                return (
                    b.material_accumulation[t, p, j] == 
                    (b.control_volume.properties_in[t].\
                     get_material_flow_terms(p, j) -
                     b.control_volume.properties_out[t].\
                         get_material_flow_terms(p, j))
                    )
            else:
                return Constraint.Skip

        # integration of material accumulation
        @self.Constraint(self.flowsheet().config.time,
                         pc_set,
                         doc="Material holdup integration")
        def material_holdup_integration(b, t, p, j):
            if (p, j) in pc_set:
                return b.material_holdup[t, p, j] == (
                      b.dt[t]
                      * b.material_accumulation[t, p, j]
                      + b.previous_material_holdup[t, p, j])

        # material holdup calculation
        @self.Constraint(self.flowsheet().config.time,
                          pc_set,
                          doc="Material holdup calculations")
        def material_holdup_calculation(b, t, p, j):
            if (p, j) in pc_set:
                return (
                    b.material_holdup[t, p, j] == (
                        b.control_volume.volume[t] *
                        b.control_volume.phase_fraction[t, p] *
                        b.control_volume.properties_out[t].\
                            get_material_density_terms(p, j)))

        # energy accumulation
        @self.Constraint(self.flowsheet().config.time,
                          doc="Energy accumulation")
        def energy_accumulation_equation(b, t):
            return (
                sum(b.energy_accumulation[t, p] for p in phase_list)
                * b.dt[t] == sum(b.energy_holdup[t, p] for p in phase_list) -
                sum(b.previous_energy_holdup[t, p] for p in phase_list)
                )

        # energy holdup calculation
        @self.Constraint(self.flowsheet().config.time,
                         phase_list,
                         doc="Energy holdup calculation")
        def energy_holdup_calculation(b, t, p):
            if (self.control_volume.properties_in[t]
                .get_material_flow_basis() == MaterialFlowBasis.molar):
                return (
                    b.energy_holdup[t, p] ==
                    (sum(b.material_holdup[t, p, j] for j in component_list)
                    * b.control_volume.properties_out[t].\
                    energy_internal_mol_phase[p])
                    )
            if (self.control_volume.properties_in[t]
                .get_material_flow_basis() == MaterialFlowBasis.mass):
                return (
                    b.energy_holdup[t, p] ==
                    (sum(b.material_holdup[t, p, j] for j in component_list)
                    * (b.control_volume.properties_out[t].\
                    energy_internal_mol_phase[p]/
                    b.control_volume.properties_out[t].mw))
                    )

        # Energy balance based on internal energy, as follows:
        #     n_final * U_final = 
        #                        n_previous * U_previous +
        #                        n_inlet * H_inlet - n_outlet * H_outlet
        #     where, n is number of moles, U is internal energy, H is enthalpy
        @self.Constraint(self.flowsheet().config.time,
                          doc="Energy balance")
        def energy_balances(b, t):
            return (
                    sum(b.energy_holdup[t, p] for p in phase_list) ==
                    sum(b.previous_energy_holdup[t, p] for p in phase_list) +
                    b.dt[t] *
                    (sum(b.control_volume.properties_in[t].
                         get_enthalpy_flow_terms(p) for p in phase_list) -
                     sum(b.control_volume.properties_out[t].
                         get_enthalpy_flow_terms(p) for p in phase_list))
                    )


    def initialize_build(blk, state_args=None, outlvl=idaeslog.NOTSET,
                   solver=None, optarg=None):
        '''
        Hydrogen tank model initialization routine.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) for the control_volume of the model to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            outlvl : sets output level of initialisation routine

                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = return solver state for each step in subroutines
                     * 3 = include solver output infomation (tee=True)

            optarg : solver options dictionary object (default={'tol': 1e-6})
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')

        Returns:
            None
        '''

        if state_args is None:
            state_args = dict()
    
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        opt = get_solver(solver, optarg)

        init_log.info_low("Starting initialization...")

        flags = blk.control_volume.initialize(state_args=state_args,
                                              outlvl=outlvl,
                                              optarg=optarg,
                                              solver=solver)
        flag_previous_state = blk.previous_state.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                hold_state=True,
                state_args=state_args,
        )
        blk.previous_state[0].sum_mole_frac_out.deactivate()

        init_log.info_high("Initialization Step 1 Complete.")

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high(
                "Initialization Step 2 {}.".format(idaeslog.condition(res))
            )
        blk.previous_state[0].sum_mole_frac_out.activate()
        blk.control_volume.release_state(flags, outlvl)
        blk.previous_state.release_state(flag_previous_state, outlvl)
        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        if not iscale.get_scaling_factor(self.control_volume.volume):
            iscale.set_scaling_factor(self.control_volume.volume,
                                      1 / value(self.tank_length[0] * ((self.tank_diameter[0]/2)**2)))

        if hasattr(self, "previous_state"):
            for t, v in self.previous_state.items():
                iscale.set_scaling_factor(v.flow_mol, 1e-3)
                iscale.set_scaling_factor(v.pressure, 1e-5)
                iscale.set_scaling_factor(v.temperature, 1e-1)

        if hasattr(self, "tank_diameter"):
            for t, v in self.tank_diameter.items():
                if not iscale.get_scaling_factor(v):
                    iscale.set_scaling_factor(v, 1)

        if hasattr(self, "tank_length"):
            for t, v in self.tank_length.items():
                if not iscale.get_scaling_factor(v):
                    iscale.set_scaling_factor(v, 1)

        if hasattr(self, "heat_duty"):
            for t, v in self.heat_duty.items():
                iscale.set_scaling_factor(v, 1e-5)

        if hasattr(self, "material_accumulation"):
            for (t, p, j), v in self.material_accumulation.items():
                if not iscale.get_scaling_factor(v):
                    iscale.set_scaling_factor(v, 1e-3)

        if hasattr(self, "energy_accumulation"):
            for (t, p), v in self.energy_accumulation.items():
                if not iscale.get_scaling_factor(v):
                    iscale.set_scaling_factor(v, 1e-3)

        if hasattr(self, "material_holdup"):
            for (t, p, j), v in self.material_holdup.items():
                if not iscale.get_scaling_factor(v):
                    iscale.set_scaling_factor(v, 1e-5)

        if hasattr(self, "energy_holdup"):
            for (t, p), v in self.energy_holdup.items():
                if not iscale.get_scaling_factor(v):
                    iscale.set_scaling_factor(v, 1e-5)

        if hasattr(self, "previous_material_holdup"):
            for (t, p, j), v in self.previous_material_holdup.items():
                if not iscale.get_scaling_factor(v):
                    iscale.set_scaling_factor(v, 1e-5)

        if hasattr(self, "previous_energy_holdup"):
            for (t, p), v in self.previous_energy_holdup.items():
                if not iscale.get_scaling_factor(v):
                    iscale.set_scaling_factor(v, 1e-5)

        # Volume constraint
        if hasattr(self, "volume_cons"):
            for t, c in self.volume_cons.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(
                        self.tank_length[t], 
                        default=1, warning=True))

        # Previous time Material Holdup Rule
        if hasattr(self, "previous_material_holdup_rule"):
            for (t, i, j), c in self.previous_material_holdup_rule.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(
                        self.material_holdup[t, i, j], 
                        default=1, warning=True))

        # Previous time Energy Holdup Rule
        if hasattr(self, "previous_energy_holdup_rule"):
            for (t, i), c in self.previous_energy_holdup_rule.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(
                        self.energy_holdup[t, i], 
                        default=1, warning=True))

        # Material Balances
        if hasattr(self, "material_balances"):
            for (t, i, j), c in self.material_balances.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(
                        self.material_accumulation[t, i, j], 
                        default=1, warning=True))

        # Material Holdup Integration
        if hasattr(self, "material_holdup_integration"):
            for (t, i, j), c in self.material_holdup_integration.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(
                        self.material_holdup[t, i, j], 
                        default=1, warning=True))

        # Material Holdup Constraints
        if hasattr(self, "material_holdup_calculation"):
            for (t, i, j), c in self.material_holdup_calculation.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(
                        self.material_holdup[t, i, j], 
                        default=1, warning=True))

        # Enthalpy Balances
        if hasattr(self, "energy_accumulation_equation"):
            for t, c in self.energy_accumulation_equation.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(
                        self.energy_accumulation[t, p], 
                        default=1, warning=True))

        # Energy Holdup Integration
        if hasattr(self, "energy_holdup_calculation"):
            for (t, i), c in self.energy_holdup_calculation.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(
                        self.energy_holdup[t, i], 
                        default=1, warning=True))

        # Energy Balance Equation
        if hasattr(self, "energy_balances"):
            for t, c in self.energy_balances.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(
                        self.energy_holdup[t, i], 
                        default=1, warning=True))

    def report(self, time_point=0, dof=False, ostream=None, prefix=""):
        super().report(time_point, dof, ostream, prefix)

        if ostream is None:
            ostream = sys.stdout

        stream_attributes = OrderedDict()
        stream_attributes["material"] = {}
        stream_attributes["energy"] = {}
        stream_attributes["material"]['initial_material_holdup'] = value(self.previous_material_holdup[0, ('Vap', 'hydrogen')])
        stream_attributes["material"]['material_holdup'] = value(self.material_holdup[0, ('Vap', 'hydrogen')])
        stream_attributes["energy"]['initial_energy_holdup'] = value(self.previous_energy_holdup[0, 'Vap'])
        stream_attributes["energy"]['energy_holdup'] = value(self.energy_holdup[0, 'Vap'])
        stream_table = DataFrame.from_dict(stream_attributes, orient="columns")

        max_str_length = 84
        tab = " " * 4

        ostream.write('\n')
        ostream.write(
            textwrap.indent(
                stream_table_dataframe_to_string(stream_table),
                prefix + tab))
        ostream.write("\n" + "=" * max_str_length + "\n")



##############################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2021 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
##############################################################################

import pandas as pd
import pytest

from pyomo.environ import (
    ConcreteModel, 
    units as pyunits, 
    value,
    assert_optimal_termination,
)
from pyomo.common import unittest as pyo_unittest
from idaes.core import FlowsheetBlock
from idaes.models.properties import iapws95
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver
from dispatches.unit_models import ConcreteTES


def _get_data():
    data = {
        "num_tubes": 10000,
        "num_segments": 20,
        "num_time_periods": 2,
        "tube_length": 64.9,
        "tube_diameter": 0.0105664,
        "face_area": 0.00847,
        "therm_cond_concrete": 1,
        "dens_mass_concrete": 2240,
        "cp_mass_concrete": 900,
        "init_temperature_concrete": [
            750, 732.631579, 715.2631579, 697.8947368, 680.5263158, 663.1578947,
            645.7894737, 628.4210526, 611.0526316, 593.6842105, 576.3157895, 558.9473684,
            541.5789474, 524.2105263, 506.8421053, 489.4736842, 472.1052632, 454.7368421,
            437.3684211, 420
        ],
        "flow_mol_charge": 0.00958 * 1000 / 18.01528,
        "inlet_pressure_charge": 19600000,
        "inlet_temperature_charge": 865,
        "flow_mol_discharge": 3 / 18.01528,
        "inlet_pressure_discharge": 8.5e5, #15 MPa
        "inlet_temperature_discharge": 355
    }

    return data


def _get_charge_results():
    exp_conc_temp = {
        1: [768.8794598487062, 750.9141725711494, 733.1558692075599, 715.5779731910243,
            698.1627726680688, 680.9003463323493, 663.7878525182592, 646.8291235216258,
            630.034517306009, 613.4209816138464, 597.0123062127739, 580.8395649489671,
            564.9418055323642, 549.3670467067806, 534.1731714688473, 519.4256478712385,
            505.4539745384297, 491.5937379825899, 477.7335015065516, 463.87326495071187],
        2: [784.6536656409681, 766.7404977929137, 749.063068065682, 731.6061482700076,
            714.3620773742523, 697.3306181729016, 680.5189998846788, 663.9421290510368,
            647.6229432955979, 631.5928719729783, 615.8923779344503, 600.5715793487628,
            585.6910142546371, 571.3226417304624, 557.5507863291356, 544.4703166829731,
            532.390904452725, 521.0060428032424, 509.9453853507483, 498.88472783457166],
    }
    exp_fluid_temp = {
        1: [843.4689736714969, 823.1455699108972, 803.8469084691471, 785.4414129181083,
            767.841394508302, 750.9977353406474, 734.896366025036, 719.5562603922092,
            705.0286981563756, 691.3975854791795, 678.7807006374081, 667.3318857141337,
            657.2444584467377, 648.7561522064175, 642.1535350190497, 637.7607287892795,
            637.2090239563571, 637.2090239563571, 637.2090239563571, 637.2090239563571],
        2: [846.9748522858338, 829.2675993812405, 811.9096875462226, 794.9307240888364,
            778.362757053882, 762.2438094603676, 746.6208988669331, 731.5526842636623,
            717.1118033575298, 703.3868998737142, 690.4843091626235, 678.5293902512656,
            667.6675857884796, 658.0654390163991, 649.9117405507793, 643.4175156823585,
            638.8141031331337, 637.2090239563571, 637.2090239563571, 637.2090239563571],
    }
    exp_heat_rate = {
        1: [581.1733601639454, 562.799805895126, 550.797916698378, 544.3495732558932,
            542.9095419858858, 546.1724178208048, 554.0507185658779, 566.6624213505045,
            584.3263730220131, 607.5642883293052, 637.1084902874657, 673.9155426951835,
            719.1874594203609, 774.4024252814344, 841.3422677079749, 922.0223200143666,
            1026.585653456652, 1134.579389291451, 1242.5731245044672, 1350.5668603392658],
        2: [485.58318181093017, 487.1875933892568, 489.6771562987536, 493.40121182477066,
            498.6691581145785, 505.77910440776037, 515.0410664427249, 526.7959467978917,
            541.4309897782801, 559.3919881808116, 581.1921956747592, 607.4178606502138,
            638.7305278065372, 675.8671612729662, 719.6417212065214, 770.9592564697472,
            829.209426341959, 905.4098030223171, 991.5902743491622, 1077.7707461721984],
    }

    return exp_conc_temp, exp_heat_rate, exp_fluid_temp


def _get_discharge_results():
    exp_conc_temp = {
        1: [746.1063169450176, 728.4696928862526, 710.5578357626713, 692.1005335939977,
            672.5608778723413, 650.8774474530392, 625.0196314618721, 592.1687287491123,
            577.7317976976101, 563.8715611417704, 550.0113246657321, 536.1510881098923,
            522.290851633854, 508.4306150780142, 494.57037860197596, 480.7101420461362,
            464.3881408074005, 446.8174177132283, 429.1096925824503, 411.20460039012323],
        2: [741.6634944375896, 723.2734354446304, 703.9106073179977, 682.6706352362868,
            657.8170956334103, 626.15173157226, 588.8776537858718, 562.6621860671783,
            551.1413184902073, 540.0806609740306, 529.0200035215366, 517.9593460053599,
            506.89868855286596, 495.83803103668924, 484.7773735841953, 473.44693663450266,
            456.95681164006436, 439.3490009451654, 421.58088889572184, 403.6002459881725],
    }
    exp_fluid_temp = {
        1: [730.7230417677312, 712.0267933383869, 691.9679135183114, 669.2086286565905,
            641.0907962507835, 602.35950271216, 542.9615404396385, 448.94200337801783,
            446.0868872570418, 446.0868872570418, 446.0868872570418, 446.0868872570418,
            446.0868872570418, 446.0868872570418, 446.0868872570418, 446.0868872570418,
            433.8991113548745, 415.5291277145009, 396.4808700496551, 376.4554822461086],
        2: [724.1106632927483, 702.7439118578757, 677.6485456948188, 645.4147211613656,
            599.5669360468476, 528.4646610349995, 446.0868872570418, 446.0868872570418,
            446.0868872570418, 446.0868872570418, 446.0868872570418, 446.0868872570418,
            446.0868872570418, 446.0868872570418, 446.0868872570418, 444.7512554661709,
            427.5969022946328, 409.8425648384156, 391.83587441116816, 373.5567435615619],
    }
    exp_heat_rate = {
        1: [-119.86067835688637, -128.11687181357422, -144.8456629119361, -178.36508955714856,
            -245.20300700827698, -378.0335268077842, -639.3656969182383, -1115.968625957053,
            -1025.7274920841446, -917.733756249346, -809.7400210363302, -701.7462852015307,
            -593.7525499885147, -485.75881415371583, -377.7650789407, -269.77134310590094,
            -237.55901851243217, -243.7865552461684, -254.23147916782654, -270.7520228978944],
        2: [-136.76504020483577, -159.9583050481918, -204.62407939245335, -290.2840313480109,
            -453.8632743728761, -761.1408110214691, -1112.5717990481025, -908.3107617856863,
            -818.544506737832, -732.3640349147958, -646.1835635879512, -560.003091764914,
            -473.82262043806975, -387.64214861503314, -301.46167728818835, -223.58592504590587,
            -228.76134048724117, -229.90302174864954, -231.7619344883195, -234.0876399576691],
    }

    return exp_conc_temp, exp_heat_rate, exp_fluid_temp


def _get_combined_results():
    exp_conc_temp = {
        1: [765.6955354841449, 747.5945530427604, 729.647450335955, 711.7483058524213,
            693.7247605780229, 675.2594659952538, 655.7351805481906, 633.9399187030289,
            607.6602996332637, 583.7078042836023, 569.918113445112, 556.5135719077973,
            543.4847736612935, 530.8394836200084, 518.5979406151248, 507.0088118612352,
            495.47770245750166, 483.64954991662637, 468.15745487706835, 451.77760745990577],
        2: [778.777670818477, 760.5255613795055, 742.4336515266298, 724.3518312101746,
            706.0253151971591, 686.9897863434737, 666.3750612481672, 642.5521353237004,
            612.6541872856708, 579.6760329417091, 566.1488472205821, 555.2224540652642,
            544.9926995318799, 535.4321187480766, 526.5379435762707, 518.6505998781274,
            510.9949538017873, 503.1420971147642, 490.91749609805186, 474.31213027291903],
    }
    exp_ch_fluid_temp = {
        1: [842.7627356797586, 821.88773432906, 802.1459839962807, 783.3568033560631,
            765.3691390138662, 748.0374738844148, 731.1872574255896, 714.5669752940619,
            697.8186750308802, 681.9136482443657, 668.4732167804137, 657.3247432678278,
            648.4122533589966, 641.7819555204223, 637.5542867815433, 637.2090239563571,
            637.2090239563571, 637.2090239563571, 637.2090239563571, 637.2090239563571],
        2: [845.6675508303674, 826.9126764153997, 808.6935241637494, 790.9589510331716,
            773.6362952280797, 756.6082192733007, 739.6737206245535, 722.493139268127,
            704.5608788442003, 686.3698150058249, 671.3190816046152, 659.2570353085273,
            649.856068774408, 642.9220654011634, 638.3585923186711, 637.2090239563571,
            637.2090239563571, 637.2090239563571, 637.2090239563571, 637.2090239563571],
    }
    exp_di_fluid_temp = {
        1: [750.6387090138653, 732.4175206365848, 713.9787935937404, 694.87288803486,
            674.2251779729404, 650.2926945836386, 619.5768887884733, 575.1169573149741,
            504.0994017758527, 446.0868872570418, 446.0868872570418, 446.0868872570418,
            446.0868872570418, 446.0868872570418, 446.0868872570418, 446.0868872570418,
            446.0868872570418, 444.3192347948272, 420.74808737074466, 391.8940779199924],
        2: [763.5730679369417, 745.2266465784485, 726.6898708802531, 707.5390824349447,
            687.0117206923077, 663.71584095563, 635.1127490396054, 596.6365248938914,
            540.4774909473294, 457.0534131014283, 446.0868872570418, 446.0868872570418,
            446.0868872570418, 446.0868872570418, 446.0868872570418, 446.0868872570418,
            446.0868872570418, 446.0868872570418, 434.5469947343781, 400.4452784614916],
    }
    exp_heat_rate = {
        1: [483.1614447654906, 460.6107363327419, 442.796967475617, 426.4595142540948,
            406.2925826295586, 372.52712256003264, 306.16235392796943, 169.88928619226263,
            -104.4274033102113, -307.1073838768562, -196.94201634240838, -74.92045305658068,
            58.66775114546273, 204.0616326381136, 361.88420465288016, 539.7902877227838,
            719.48239872132, 890.0305245956872, 947.7901571239249, 978.2217845978254],
        2: [402.71218622254713, 398.05998824109133, 393.6023365745825, 387.9789591850555,
            378.65249944496645, 361.098768037143, 327.5313630634786, 265.11303352984964,
            153.72868135470637, -124.11150090935013, -116.03070927967546, -39.744955678063576,
            46.419036990981446, 141.37677723704962, 244.42003307404175, 358.3734447019817,
            477.6732584796396, 600.046253657646, 700.6307246007051, 693.6885084573833],
    }

    return exp_conc_temp, exp_heat_rate, exp_ch_fluid_temp, exp_di_fluid_temp


@pytest.fixture(scope="module")
def build_concrete_tes_charge():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic="False")
    m.fs.props_water = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.LG,
    )

    data = _get_data()
    inlet_enthalpy_charge = iapws95.htpx(
        T=data["inlet_temperature_charge"] * pyunits.K,
        P=data["inlet_pressure_charge"] * pyunits.Pa,
    )

    m.fs.tes = ConcreteTES(
        model_data=data, 
        property_package=m.fs.props_water,
        operating_mode="charge",
    )

    m.fs.tes.inlet_charge.flow_mol.fix(data["flow_mol_charge"] * data["num_tubes"])
    m.fs.tes.inlet_charge.pressure.fix(data["inlet_pressure_charge"])
    m.fs.tes.inlet_charge.enth_mol.fix(inlet_enthalpy_charge)

    return m


@pytest.fixture(scope="module")
def build_concrete_tes_discharge():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic="False")
    m.fs.props_water = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.LG,
    )

    data = _get_data()
    inlet_enthalpy_discharge = iapws95.htpx(
        T=data["inlet_temperature_discharge"] * pyunits.K,
        P=data["inlet_pressure_discharge"] * pyunits.Pa,
    )

    m.fs.tes = ConcreteTES(
        model_data=data, 
        property_package=m.fs.props_water,
        operating_mode="discharge",
    )

    m.fs.tes.inlet_discharge.flow_mol.fix(data["flow_mol_discharge"] * data["num_tubes"])
    m.fs.tes.inlet_discharge.pressure.fix(data["inlet_pressure_discharge"])
    m.fs.tes.inlet_discharge.enth_mol.fix(inlet_enthalpy_discharge)

    return m


@pytest.fixture(scope="module")
def build_concrete_tes_charge_discharge():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic="False")
    m.fs.props_water = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.LG,
    )

    data = _get_data()
    inlet_enthalpy_charge = iapws95.htpx(
        T=data["inlet_temperature_charge"] * pyunits.K,
        P=data["inlet_pressure_charge"] * pyunits.Pa,
    )
    inlet_enthalpy_discharge = iapws95.htpx(
        T=data["inlet_temperature_discharge"] * pyunits.K,
        P=data["inlet_pressure_discharge"] * pyunits.Pa,
    )

    m.fs.tes = ConcreteTES(
        model_data=data, 
        property_package=m.fs.props_water,
        operating_mode="combined",
    )

    m.fs.tes.inlet_charge.flow_mol.fix(data["flow_mol_charge"] * data["num_tubes"])
    m.fs.tes.inlet_charge.pressure.fix(data["inlet_pressure_charge"])
    m.fs.tes.inlet_charge.enth_mol.fix(inlet_enthalpy_charge)

    m.fs.tes.inlet_discharge.flow_mol.fix((0.01 / 3) * data["flow_mol_discharge"] * data["num_tubes"])
    m.fs.tes.inlet_discharge.pressure.fix(data["inlet_pressure_discharge"])
    m.fs.tes.inlet_discharge.enth_mol.fix(inlet_enthalpy_discharge)

    return m


@pytest.fixture(scope="module")
def build_concrete_tes_discharge_charge():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic="False")
    m.fs.props_water = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.LG,
    )

    data = _get_data()
    inlet_enthalpy_charge = iapws95.htpx(
        T=data["inlet_temperature_charge"] * pyunits.K,
        P=data["inlet_pressure_charge"] * pyunits.Pa,
    )
    inlet_enthalpy_discharge = iapws95.htpx(
        T=data["inlet_temperature_discharge"] * pyunits.K,
        P=data["inlet_pressure_discharge"] * pyunits.Pa,
    )

    m.fs.tes = ConcreteTES(
        model_data=data, 
        property_package=m.fs.props_water,
        operating_mode="combined",
    )

    m.fs.tes.inlet_charge.flow_mol.fix((0.001 / 9.58) * data["flow_mol_charge"] * data["num_tubes"])
    m.fs.tes.inlet_charge.pressure.fix(data["inlet_pressure_charge"])
    m.fs.tes.inlet_charge.enth_mol.fix(inlet_enthalpy_charge)

    m.fs.tes.inlet_discharge.flow_mol.fix(data["flow_mol_discharge"] * data["num_tubes"])
    m.fs.tes.inlet_discharge.pressure.fix(data["inlet_pressure_discharge"])
    m.fs.tes.inlet_discharge.enth_mol.fix(inlet_enthalpy_discharge)

    return m


@pytest.fixture(scope="module")
def build_concrete_tes_combined():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic="False")
    m.fs.props_water = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.LG,
    )

    data = _get_data()
    inlet_enthalpy_charge = iapws95.htpx(
        T=data["inlet_temperature_charge"] * pyunits.K,
        P=data["inlet_pressure_charge"] * pyunits.Pa,
    )
    inlet_enthalpy_discharge = iapws95.htpx(
        T=data["inlet_temperature_discharge"] * pyunits.K,
        P=data["inlet_pressure_discharge"] * pyunits.Pa,
    )

    m.fs.tes = ConcreteTES(
        model_data=data, 
        property_package=m.fs.props_water,
        operating_mode="combined",
    )

    m.fs.tes.inlet_charge.flow_mol.fix(data["flow_mol_charge"] * data["num_tubes"])
    m.fs.tes.inlet_charge.pressure.fix(data["inlet_pressure_charge"])
    m.fs.tes.inlet_charge.enth_mol.fix(inlet_enthalpy_charge)

    m.fs.tes.inlet_discharge.flow_mol.fix(data["flow_mol_discharge"] * data["num_tubes"])
    m.fs.tes.inlet_discharge.pressure.fix(data["inlet_pressure_discharge"])
    m.fs.tes.inlet_discharge.enth_mol.fix(inlet_enthalpy_discharge)

    return m


@pytest.mark.unit
def test_tes_inputs():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic="False")
    m.fs.props_water = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.LG,
    )

    data = _get_data()
    data.pop("num_segments")

    with pytest.raises(
        KeyError,
        match=(
            f"Property num_segments, a required argument, is not in model_data"
        ),
    ):
        m.fs.tes = ConcreteTES(
            model_data=data, 
            property_package=m.fs.props_water,
            operating_mode="charge",
        )


@pytest.mark.unit
def test_tes_charge(build_concrete_tes_charge):
    m = build_concrete_tes_charge

    assert len(m.fs.tes.time_periods) == 2
    for t in m.fs.tes.time_periods:
        assert hasattr(m.fs.tes.period[t], "concrete")
        assert hasattr(m.fs.tes.period[t], "tube_charge")
        assert not hasattr(m.fs.tes.period[t], "tube_discharge")

    m.fs.tes.initialize()

    assert degrees_of_freedom(m) == 0

    solver = get_solver()
    result = solver.solve(m, tee=True)

    assert_optimal_termination(result)

    conc_temp_prof = {}
    fluid_temp_prof = {}
    heat_rate_prof = {}

    exp_conc_temp, exp_heat_rate, exp_fluid_temp = _get_charge_results()

    for t in m.fs.tes.period:
        conc_temp_prof[t] = [
            value(m.fs.tes.period[t].concrete.temperature[s]) for s in m.fs.tes.segments
        ]
        heat_rate_prof[t] = [
            value(m.fs.tes.period[t].concrete.heat_rate[s]) for s in m.fs.tes.segments
        ]
        fluid_temp_prof[t] = [
            value(m.fs.tes.period[t].tube_charge.hex[s].control_volume.properties_out[0].temperature)
            for s in m.fs.tes.segments
        ]
        # NS: added abstol=1 in order to address the test failures due to triggered default reltol = 1e-7 
        pyo_unittest.assertStructuredAlmostEqual(first=exp_conc_temp[t], second=conc_temp_prof[t], abstol=1)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_fluid_temp[t], second=fluid_temp_prof[t], abstol=1)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_heat_rate[t], second=heat_rate_prof[t], abstol=1)
    
    return


@pytest.mark.unit
def test_tes_discharge(build_concrete_tes_discharge):
    m = build_concrete_tes_discharge

    assert len(m.fs.tes.time_periods) == 2
    for t in m.fs.tes.time_periods:
        assert hasattr(m.fs.tes.period[t], "concrete")
        assert not hasattr(m.fs.tes.period[t], "tube_charge")
        assert hasattr(m.fs.tes.period[t], "tube_discharge")

    m.fs.tes.initialize()

    assert degrees_of_freedom(m) == 0

    solver = get_solver()
    result = solver.solve(m, tee=True)

    assert_optimal_termination(result)

    conc_temp_prof = {}
    fluid_temp_prof = {}
    heat_rate_prof = {}

    exp_conc_temp, exp_heat_rate, exp_fluid_temp = _get_discharge_results()

    for t in m.fs.tes.period:
        conc_temp_prof[t] = [
            value(m.fs.tes.period[t].concrete.temperature[s]) for s in m.fs.tes.segments
        ]
        heat_rate_prof[t] = [
            value(m.fs.tes.period[t].concrete.heat_rate[s]) for s in m.fs.tes.segments
        ]
        fluid_temp_prof[t] = [
            value(m.fs.tes.period[t].tube_discharge.hex[s].control_volume.properties_out[0].temperature)
            for s in m.fs.tes.segments
        ]
        # NS: added abstol=1 in order to address the test failures due to triggered default reltol = 1e-7 
        pyo_unittest.assertStructuredAlmostEqual(first=exp_conc_temp[t], second=conc_temp_prof[t], abstol=1)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_fluid_temp[t], second=fluid_temp_prof[t], abstol=1)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_heat_rate[t], second=heat_rate_prof[t], abstol=1)
    
    return


@pytest.mark.unit
def test_tes_charge_discharge(build_concrete_tes_charge_discharge):
    m = build_concrete_tes_charge_discharge

    assert len(m.fs.tes.time_periods) == 2
    for t in m.fs.tes.time_periods:
        assert hasattr(m.fs.tes.period[t], "concrete")
        assert hasattr(m.fs.tes.period[t], "tube_charge")
        assert hasattr(m.fs.tes.period[t], "tube_discharge")

    m.fs.tes.initialize()

    assert degrees_of_freedom(m) == 0

    solver = get_solver()
    result = solver.solve(m, tee=True)

    assert_optimal_termination(result)

    conc_temp_prof = {}
    ch_fluid_temp_prof = {}
    di_fluid_temp_prof = {}
    heat_rate_prof = {}

    exp_conc_temp, exp_heat_rate, exp_fluid_temp = _get_charge_results()

    for t in m.fs.tes.period:
        conc_temp_prof[t] = [
            value(m.fs.tes.period[t].concrete.temperature[s]) for s in m.fs.tes.segments
        ]
        heat_rate_prof[t] = [
            value(m.fs.tes.period[t].concrete.heat_rate[s]) for s in m.fs.tes.segments
        ]
        ch_fluid_temp_prof[t] = [
            value(m.fs.tes.period[t].tube_charge.hex[s].control_volume.properties_out[0].temperature)
            for s in m.fs.tes.segments
        ]
        di_fluid_temp_prof[t] = [
            value(m.fs.tes.period[t].tube_discharge.hex[s].control_volume.properties_out[0].temperature)
            for s in m.fs.tes.segments
        ]

        pyo_unittest.assertStructuredAlmostEqual(first=exp_conc_temp[t], second=conc_temp_prof[t], abstol=5)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_fluid_temp[t], second=ch_fluid_temp_prof[t], abstol=5)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_heat_rate[t], second=heat_rate_prof[t], abstol=30)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_conc_temp[t], second=di_fluid_temp_prof[t], abstol=10)
    
    return


@pytest.mark.unit
def tes_charge_discharge_test(build_concrete_tes_discharge_charge):
    m = build_concrete_tes_discharge_charge

    assert len(m.fs.tes.time_periods) == 2
    for t in m.fs.tes.time_periods:
        assert hasattr(m.fs.tes.period[t], "concrete")
        assert hasattr(m.fs.tes.period[t], "tube_charge")
        assert hasattr(m.fs.tes.period[t], "tube_discharge")

    m.fs.tes.initialize()

    assert degrees_of_freedom(m) == 0

    solver = get_solver()
    result = solver.solve(m, tee=True)

    assert_optimal_termination(result)

    conc_temp_prof = {}
    ch_fluid_temp_prof = {}
    di_fluid_temp_prof = {}
    heat_rate_prof = {}

    exp_conc_temp, exp_heat_rate, exp_fluid_temp = _get_discharge_results()

    for t in m.fs.tes.period:
        conc_temp_prof[t] = [
            value(m.fs.tes.period[t].concrete.temperature[s]) for s in m.fs.tes.segments
        ]
        heat_rate_prof[t] = [
            value(m.fs.tes.period[t].concrete.heat_rate[s]) for s in m.fs.tes.segments
        ]
        ch_fluid_temp_prof[t] = [
            value(m.fs.tes.period[t].tube_charge.hex[s].control_volume.properties_out[0].temperature)
            for s in m.fs.tes.segments
        ]
        di_fluid_temp_prof[t] = [
            value(m.fs.tes.period[t].tube_discharge.hex[s].control_volume.properties_out[0].temperature)
            for s in m.fs.tes.segments
        ]

        pyo_unittest.assertStructuredAlmostEqual(first=exp_conc_temp[t], second=conc_temp_prof[t], abstol=5)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_conc_temp[t], second=ch_fluid_temp_prof[t], abstol=5)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_heat_rate[t], second=heat_rate_prof[t], abstol=30)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_fluid_temp[t], second=di_fluid_temp_prof[t], abstol=5)
    
    return


@pytest.mark.unit
def test_tes_combined(build_concrete_tes_combined):
    m = build_concrete_tes_combined

    assert len(m.fs.tes.time_periods) == 2
    for t in m.fs.tes.time_periods:
        assert hasattr(m.fs.tes.period[t], "concrete")
        assert hasattr(m.fs.tes.period[t], "tube_charge")
        assert hasattr(m.fs.tes.period[t], "tube_discharge")

    m.fs.tes.initialize()

    assert degrees_of_freedom(m) == 0

    solver = get_solver()
    result = solver.solve(m, tee=True)

    assert_optimal_termination(result)

    conc_temp_prof = {}
    ch_fluid_temp_prof = {}
    di_fluid_temp_prof = {}
    heat_rate_prof = {}

    exp_conc_temp, exp_heat_rate, exp_ch_fluid_temp, exp_di_fluid_temp = _get_combined_results()

    for t in m.fs.tes.period:
        conc_temp_prof[t] = [
            value(m.fs.tes.period[t].concrete.temperature[s]) for s in m.fs.tes.segments
        ]
        heat_rate_prof[t] = [
            value(m.fs.tes.period[t].concrete.heat_rate[s]) for s in m.fs.tes.segments
        ]
        ch_fluid_temp_prof[t] = [
            value(m.fs.tes.period[t].tube_charge.hex[s].control_volume.properties_out[0].temperature)
            for s in m.fs.tes.segments
        ]
        di_fluid_temp_prof[t] = [
            value(m.fs.tes.period[t].tube_discharge.hex[s].control_volume.properties_out[0].temperature)
            for s in m.fs.tes.segments
        ]
        # NS: added abstol=1 in order to address the test failures due to triggered default reltol = 1e-7 
        pyo_unittest.assertStructuredAlmostEqual(first=exp_conc_temp[t], second=conc_temp_prof[t], abstol=1)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_ch_fluid_temp[t], second=ch_fluid_temp_prof[t], abstol=1)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_heat_rate[t], second=heat_rate_prof[t], abstol=1)
        pyo_unittest.assertStructuredAlmostEqual(first=exp_di_fluid_temp[t], second=di_fluid_temp_prof[t], abstol=1)
    
    return



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################

"""

This is a simple model for an ultra-supercritical pulverized coal power plant
based on a flowsheet presented in Ref [1]: 1999 USDOE Report #DOE/FE-0400.
This model uses some of the simpler unit models from the power generation
unit model library and some of the parameters in the model,
such as feed water heater areas, overall heat
transfer coefficients, and turbine efficiencies at multiple stages
have all been estimated for a total power out of 437 MW.
Additional assumptions are as follows:
(1) The flowsheet and main steam conditions, i. e. pressure & temperature
are adopted from the aforementioned DOE report
(2) Heater unit models are used to model main steam boiler, reheater,
and condenser.
(3) Multi-stage turbines are modeled as multiple lumped single
stage turbines
"""

__author__ = "Naresh Susarla & E S Rawlings"

import os
import logging

# Import Pyomo libraries
from pyomo.environ import (ConcreteModel, RangeSet, TransformationFactory,
                           Constraint, Param, Var, Reals, value)
from pyomo.environ import units as pyunits
from pyomo.network import Arc
from pyomo.common.fileutils import this_file_dir

# Import IDAES libraries
from idaes.core import FlowsheetBlock, MaterialBalanceType
from idaes.core.solvers.get_solver import get_solver
from idaes.core.util.initialization import propagate_state
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models import (
    HeatExchanger,
    MomentumMixingType,
    Heater,
)
from idaes.models_extra.power_generation.unit_models.helm import (
    HelmMixer,
    HelmIsentropicCompressor,
    HelmTurbineStage,
    HelmSplitter
)
from idaes.models.unit_models.heat_exchanger import (
    delta_temperature_underwood_callback)
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.util.tags import svg_tag, ModelTagGroup

# Import Property Packages (IAPWS95 for Water/Steam)
from idaes.models.properties import iapws95


def declare_unit_model(m=None):
    """Create flowsheet and add unit models.
    """
    ###########################################################################
    #  Flowsheet and Property Package                                         #
    ###########################################################################
    if m is None:
        m = ConcreteModel(name="Ultra Supercritical Power Plant Model")

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.prop_water = iapws95.Iapws95ParameterBlock()

    ###########################################################################
    #   Turbine declarations                                   #
    ###########################################################################
    # A total of 11 single stage turbines are used to model
    # different multistage turbines
    m.set_turbine = RangeSet(11)
    m.fs.turbine = HelmTurbineStage(
        m.set_turbine,
        property_package=m.fs.prop_water,
    )

    #########################################################################
    #  Turbine outlet splitters                                  #
    #########################################################################
    # The default number of outlets for a splitter is 2. This can be changed
    # using the "num_outlets" argument.
    # In the USC flowsheet turbine_splitter[6] has 3 outlets. This is realized
    # by using the 'initialize' argument as shown below.
    m.set_turbine_splitter = RangeSet(10)
    m.fs.turbine_splitter = HelmSplitter(
        m.set_turbine_splitter,
        property_package=m.fs.prop_water,
        initialize={
            6: {
                "property_package": m.fs.prop_water,
                "num_outlets": 3
                },
        }
    )

    ###########################################################################
    #  Boiler section & condenser declarations:                               #
    ###########################################################################
    # Boiler section is set up using three heater blocks, as following:
    # 1) For the main steam the heater block is named 'boiler'
    # 2) For the 1st reheated steam the heater block is named 'reheater_1'
    # 3) For the 2nd reheated steam the heater block is named 'reheater_2'

    m.fs.boiler = Heater(
        property_package=m.fs.prop_water,
        has_pressure_change=True,
    )
    # Set and indexed units for reheater
    m.set_reheater = RangeSet(2)
    m.fs.reheater = Heater(
        m.set_reheater,
        property_package=m.fs.prop_water,
        has_pressure_change=True,
    )

    ###########################################################################
    #  Add Condenser Mixer, Condenser, and Condensate pump                    #
    ###########################################################################
    # condenser mix
    # The inlet 'main' refers to the main steam coming from the turbine train
    # Inlet 'bfpt' refers to the steam coming from the bolier feed pump turbine
    # Inlet 'drain' refers to the condensed steam from the feed water heater 1
    # Inlet 'makeup' refers to the make up water
    m.fs.condenser_mix = HelmMixer(
        momentum_mixing_type=MomentumMixingType.minimize,
        inlet_list=["main", "bfpt", "drain", "makeup"],
        property_package=m.fs.prop_water,
    )

    # Condenser is set up as a heater block
    m.fs.condenser = Heater(
        property_package=m.fs.prop_water,
        has_pressure_change=False,
    )

    # condensate pump
    m.fs.cond_pump = HelmIsentropicCompressor(
        property_package=m.fs.prop_water,
    )
    ###########################################################################
    #  Feedwater heater declaration                                     #
    ###########################################################################
    # Feed water heaters (FWHs) are declared as 0D heat exchangers
    # Shell side (side 1) is for steam condensing
    # Tube side (side 2) is for feed water heating

    # Declaring indexed units for feed water heater mixers
    # The indices reflect the connection to corresponding feed water heaters
    # e. g. the outlet of fwh_mixer[1] is connected to fwh[1]
    # Note that there are no mixers before FWHs 5 and 9
    m.set_fwh_mixer = [1, 2, 3, 4, 6, 7, 8]
    m.fs.fwh_mixer = HelmMixer(
        m.set_fwh_mixer,
        momentum_mixing_type=MomentumMixingType.minimize,
        inlet_list=["steam", "drain"],
        property_package=m.fs.prop_water,
    )

    # Declaring  set and indexed units for feed water heaters
    m.set_fwh = RangeSet(9)
    m.fs.fwh = HeatExchanger(
        m.set_fwh,
        delta_temperature_callback=delta_temperature_underwood_callback,
        hot_side_name="shell",
        cold_side_name="tube",
        shell={
            "property_package": m.fs.prop_water,
            "material_balance_type": MaterialBalanceType.componentTotal,
            "has_pressure_change": True,
        },
        tube={
            "property_package": m.fs.prop_water,
            "material_balance_type": MaterialBalanceType.componentTotal,
            "has_pressure_change": True,
        },
    )

    ###########################################################################
    #  Add deaerator                              #
    ###########################################################################
    m.fs.deaerator = HelmMixer(
        momentum_mixing_type=MomentumMixingType.minimize,
        inlet_list=["steam", "drain", "feedwater"],
        property_package=m.fs.prop_water,
    )

    ###########################################################################
    #  Add auxiliary turbine, booster pump, and boiler feed pump (BFP)        #
    ###########################################################################
    m.fs.booster = HelmIsentropicCompressor(
        property_package=m.fs.prop_water,
    )
    m.fs.bfp = HelmIsentropicCompressor(
        property_package=m.fs.prop_water,
    )
    m.fs.bfpt = HelmTurbineStage(
            property_package=m.fs.prop_water,
    )

    ###########################################################################
    #  Create the stream Arcs and return the model                            #
    ###########################################################################
    _make_constraints(m)
    _create_arcs(m)
    TransformationFactory("network.expand_arcs").apply_to(m.fs)
    return m


def _make_constraints(m):
    # Define all model constraints except those included with the unit models

    # ********* Boiler section *********
    # Following the Ref [1], the outlet temperature for the boiler units,
    # i.e., boiler and reheaters [1] & [2], are fixed to 866.15 K
    def temperature_constraint(b, t):
        return (b.control_volume.properties_out[t].temperature ==
                866.15 * pyunits.K)

    for unit in [m.fs.boiler, m.fs.reheater[1], m.fs.reheater[2]]:
        setattr(unit,
                "temperature_constraint",
                Constraint(m.fs.config.time,
                           rule=temperature_constraint))

    # ********* Condenser section *********
    # The outlet of condenser is assumed to be a saturated liquid
    # The following condensate enthalpy at the outlet of condeser equal to
    # that of a saturated liquid at that pressure
    @m.fs.condenser.Constraint(m.fs.time)
    def cond_vaporfrac_constraint(b, t):
        return (
            b.control_volume.properties_out[t].enth_mol
            == b.control_volume.properties_out[t].enth_mol_sat_phase['Liq']
        )

    # ********* Feed Water Heater section *********
    # The condensing steam is assumed to leave the FWH as saturated liquid
    # Thus, each FWH is accompanied by 3 constraints, 2 for pressure drop
    # and 1 for the enthalpy.
    # Side 1 outlet of fwh is assumed to be a saturated liquid
    # The following constraint sets the side 1 outlet enthalpy to be
    # same as that of saturated liquid
    def fwh_vaporfrac_constraint(b, t):
        return (
            b.hot_side.properties_out[t].enth_mol ==
            b.hot_side.properties_out[t].enth_mol_sat_phase['Liq']
        )
    for i in m.set_fwh:
        b = m.fs.fwh[i]
        setattr(b,
                "fwh_vfrac_constraint",
                Constraint(m.fs.config.time,
                           rule=fwh_vaporfrac_constraint))

    # Pressure drop on both sides are accounted for by setting the respective
    # outlet pressure based on the following assumptions:
    #     (1) Feed water side (side 2): A constant 4% pressure drop is assumed
    #           on the feedwater side for all FWHs. For this,
    #           the outlet pressure is set to 0.96 times the inlet pressure,
    #           on the feed water side for all FWHs
    #     (2) Steam condensing side (side 1): Going from high pressure to
    #           low pressure FWHs, the outlet pressure of
    #           the condensed steam in assumed to be 10% more than that
    #           of the pressure of steam extracted for the immediately
    #           next lower pressure feedwater heater.
    #           e.g. the outlet condensate pressure of FWH 'n'
    #           = 1.1 * pressure of steam extracted for FWH 'n-1'
    #           In case of FWH1 the FWH 'n-1' is used for Condenser,
    #           and in case of FWH6, FWH 'n-1' is for Deaerator. Here,
    #           the steam pressure for FWH 'n-1' is known because the
    #           pressure ratios for turbines are fixed.

    # Side 2 pressure drop constraint
    # Setting a 4% pressure drop on the feedwater side (P_out = 0.96 * P_in)
    def fwh_s2pdrop_constraint(b, t):
        return (
            b.cold_side.properties_out[t].pressure ==
            0.96 * b.cold_side.properties_in[t].pressure
        )
    for i in m.set_fwh:
        b = m.fs.fwh[i]
        setattr(b,
                "fwh_s2_delp_constraint",
                Constraint(m.fs.config.time,
                           rule=fwh_s2pdrop_constraint))

    # Side 1 pressure drop constraint
    # Setting the outlet pressure as described above. For this, the
    # relevant turbine stage pressure ratios are used (pressure_ratio_dict)
    # The pressure drop across the reheaters 1 and 2 are also accounted
    # in case of fwh[9] and fwh[7], respectively
    # For this, pressure_diffr_dict is defined

    # 0.204 is the pressure ratio for turbine #11 (see set_inputs)
    # 0.476 is the pressure ratio for turbine #10 (see set_inputs)
    # 0.572 is the pressure ratio for turbine #9 (see set_inputs)
    # 0.389 is the pressure ratio for turbine #8 (see set_inputs)
    # 0.514 is the pressure ratio for turbine #7 (see set_inputs)
    # 0.523 is the pressure ratio for turbine #5 (see set_inputs)
    # 0.609 is the pressure ratio for turbine #4 (see set_inputs)
    # 0.498 is the pressure ratio for turbine #3 (see set_inputs)
    # 0.774 is the pressure ratio for turbine #2 (see set_inputs)
    m.data_pressure_ratio = {1: 0.204,
                             2: 0.476,
                             3: 0.572,
                             4: 0.389,
                             5: 0.514,
                             6: 0.523,
                             7: 0.609,
                             8: 0.498,
                             9: 0.774}

    # 742845 Pa is the pressure drop across reheater_1
    # 210952 Pa is the pressure drop across reheater_2
    m.data_pressure_diffr = {1: 0,
                             2: 0,
                             3: 0,
                             4: 0,
                             5: 0,
                             6: 210952,
                             7: 0,
                             8: 742845,
                             9: 0}

    def fwh_s1pdrop_constraint(b, t):
        return (
            b.hot_side.properties_out[t].pressure ==
            1.1 * b.turb_press_ratio *
            (b.hot_side.properties_in[t].pressure - b.reheater_press_diff)
        )

    for i in m.set_fwh:
        b = m.fs.fwh[i]
        b.turb_press_ratio = Param(initialize=m.data_pressure_ratio[i],
                                   units=pyunits.Pa/pyunits.Pa)
        b.reheater_press_diff = Param(initialize=m.data_pressure_diffr[i],
                                      units=pyunits.Pa)
        setattr(b, "s1_delp_constraint",
                Constraint(m.fs.config.time, rule=fwh_s1pdrop_constraint))

    # The following constraint sets the outlet pressure of steam extracted
    # for boiler feed water turbine to be same as that of condenser
    @m.fs.Constraint(m.fs.time)
    def constraint_out_pressure(b, t):
        return (
            b.bfpt.control_volume.properties_out[t].pressure ==
            b.condenser_mix.main_state[t].pressure
        )

    # The following constraint demands that the work done by the
    # boiler feed water pump is same as that of boiler feed water turbine
    # Essentially, this says that boiler feed water turbine produces just
    # enough power to meet the demand of boiler feed water pump
    @m.fs.Constraint(m.fs.time)
    def constraint_bfp_power(b, t):
        return (
            b.booster.control_volume.work[t] +
            b.bfp.control_volume.work[t] +
            b.bfpt.control_volume.work[t] +
            b.cond_pump.control_volume.work[t] ==
            0
        )

    # The power plant with storage for a charge scenario is now ready
    # Declaraing variables for plant power out and plant heat duty
    # for use in analysis of various design and operating scenarios
    m.fs.plant_power_out = Var(
        m.fs.time,
        domain=Reals,
        initialize=400,
        doc="Net Power MWe out from the power plant",
        units=pyunits.MW
    )
    m.fs.plant_heat_duty = Var(
        m.fs.time,
        domain=Reals,
        initialize=400,
        doc="Net Power MWe out from the power plant",
        units=pyunits.MW
    )

    #   Constraint on Plant Power Output
    #   Plant Power Out = Total Turbine Power
    @m.fs.Constraint(m.fs.time)
    def production_cons(b, t):
        return (
            (-1*sum(m.fs.turbine[p].work_mechanical[t]
                    for p in m.set_turbine)) ==
            m.fs.plant_power_out[t]*1e6*(pyunits.W/pyunits.MW)
        )

    #   Constraint on Plant Power Output
    #   Plant Power Out = Total Turbine Power
    @m.fs.Constraint(m.fs.time)
    def heatduty_cons(b, t):
        return (
            (sum(unit.heat_duty[t]
                 for unit in [m.fs.boiler, m.fs.reheater[1], m.fs.reheater[2]])
             ) ==
            m.fs.plant_heat_duty[t]*1e6*(pyunits.W/pyunits.MW)
        )


def _create_arcs(m):

    # boiler to turb
    m.fs.boiler_to_turb1 = Arc(
        source=m.fs.boiler.outlet, destination=m.fs.turbine[1].inlet
    )

    # turbine1 splitter
    m.fs.turb1_to_t1split = Arc(
        source=m.fs.turbine[1].outlet,
        destination=m.fs.turbine_splitter[1].inlet
    )
    m.fs.t1split_to_turb2 = Arc(
        source=m.fs.turbine_splitter[1].outlet_1,
        destination=m.fs.turbine[2].inlet
    )
    m.fs.t1split_to_fwh9 = Arc(
        source=m.fs.turbine_splitter[1].outlet_2,
        destination=m.fs.fwh[9].shell_inlet
    )

    # turbine2 splitter
    m.fs.turb2_to_t2split = Arc(
        source=m.fs.turbine[2].outlet,
        destination=m.fs.turbine_splitter[2].inlet
    )
    m.fs.t2split_to_rh1 = Arc(
        source=m.fs.turbine_splitter[2].outlet_1,
        destination=m.fs.reheater[1].inlet
    )
    m.fs.t2split_to_fwh8mix = Arc(
        source=m.fs.turbine_splitter[2].outlet_2,
        destination=m.fs.fwh_mixer[8].steam
    )

    # reheater_1 to turbine_3
    m.fs.rh1_to_turb3 = Arc(
        source=m.fs.reheater[1].outlet, destination=m.fs.turbine[3].inlet
    )

    # turbine3 splitter
    m.fs.turb3_to_t3split = Arc(
        source=m.fs.turbine[3].outlet,
        destination=m.fs.turbine_splitter[3].inlet
    )
    m.fs.t3split_to_turb4 = Arc(
        source=m.fs.turbine_splitter[3].outlet_1,
        destination=m.fs.turbine[4].inlet
    )
    m.fs.t3split_to_fwh7mix = Arc(
        source=m.fs.turbine_splitter[3].outlet_2,
        destination=m.fs.fwh_mixer[7].steam
    )

    # turbine4 splitter
    m.fs.turb4_to_t4split = Arc(
        source=m.fs.turbine[4].outlet,
        destination=m.fs.turbine_splitter[4].inlet
    )
    m.fs.t4split_to_rh2 = Arc(
        source=m.fs.turbine_splitter[4].outlet_1,
        destination=m.fs.reheater[2].inlet
    )
    m.fs.t4split_to_fwh6mix = Arc(
        source=m.fs.turbine_splitter[4].outlet_2,
        destination=m.fs.fwh_mixer[6].steam
    )

    # reheater_2 to turbine_5
    m.fs.rh2_to_turb5 = Arc(
        source=m.fs.reheater[2].outlet, destination=m.fs.turbine[5].inlet
    )

    # turbine5 splitter
    m.fs.turb5_to_t5split = Arc(
        source=m.fs.turbine[5].outlet,
        destination=m.fs.turbine_splitter[5].inlet
    )
    m.fs.t5split_to_turb6 = Arc(
        source=m.fs.turbine_splitter[5].outlet_1,
        destination=m.fs.turbine[6].inlet
    )
    m.fs.t5split_to_deaerator = Arc(
        source=m.fs.turbine_splitter[5].outlet_2,
        destination=m.fs.deaerator.steam
    )

    # turbine6 splitter
    m.fs.turb6_to_t6split = Arc(
        source=m.fs.turbine[6].outlet,
        destination=m.fs.turbine_splitter[6].inlet
    )
    m.fs.t6split_to_turb7 = Arc(
        source=m.fs.turbine_splitter[6].outlet_1,
        destination=m.fs.turbine[7].inlet
    )
    m.fs.t6split_to_fwh5 = Arc(
        source=m.fs.turbine_splitter[6].outlet_2,
        destination=m.fs.fwh[5].shell_inlet
    )
    m.fs.t6split_to_bfpt = Arc(
        source=m.fs.turbine_splitter[6].outlet_3,
        destination=m.fs.bfpt.inlet
    )

    # turbine7 splitter
    m.fs.turb7_to_t7split = Arc(
        source=m.fs.turbine[7].outlet,
        destination=m.fs.turbine_splitter[7].inlet
    )
    m.fs.t7split_to_turb8 = Arc(
        source=m.fs.turbine_splitter[7].outlet_1,
        destination=m.fs.turbine[8].inlet
    )
    m.fs.t7split_to_fwh4mix = Arc(
        source=m.fs.turbine_splitter[7].outlet_2,
        destination=m.fs.fwh_mixer[4].steam
    )

    # turbine8 splitter
    m.fs.turb8_to_t8split = Arc(
        source=m.fs.turbine[8].outlet,
        destination=m.fs.turbine_splitter[8].inlet
    )
    m.fs.t8split_to_turb9 = Arc(
        source=m.fs.turbine_splitter[8].outlet_1,
        destination=m.fs.turbine[9].inlet
    )
    m.fs.t8split_to_fwh3mix = Arc(
        source=m.fs.turbine_splitter[8].outlet_2,
        destination=m.fs.fwh_mixer[3].steam
    )

    # turbine9 splitter
    m.fs.turb9_to_t9split = Arc(
        source=m.fs.turbine[9].outlet,
        destination=m.fs.turbine_splitter[9].inlet
    )
    m.fs.t9split_to_turb10 = Arc(
        source=m.fs.turbine_splitter[9].outlet_1,
        destination=m.fs.turbine[10].inlet
    )
    m.fs.t9split_to_fwh2mix = Arc(
        source=m.fs.turbine_splitter[9].outlet_2,
        destination=m.fs.fwh_mixer[2].steam
    )

    # turbine10 splitter
    m.fs.turb10_to_t10split = Arc(
        source=m.fs.turbine[10].outlet,
        destination=m.fs.turbine_splitter[10].inlet
    )
    m.fs.t10split_to_turb11 = Arc(
        source=m.fs.turbine_splitter[10].outlet_1,
        destination=m.fs.turbine[11].inlet
    )
    m.fs.t10split_to_fwh1mix = Arc(
        source=m.fs.turbine_splitter[10].outlet_2,
        destination=m.fs.fwh_mixer[1].steam
    )

    # condenser mixer to condensate pump
    m.fs.turb11_to_condmix = Arc(
        source=m.fs.turbine[11].outlet,
        destination=m.fs.condenser_mix.main
    )
    m.fs.fwh1_to_condmix = Arc(
        source=m.fs.fwh[1].shell_outlet,
        destination=m.fs.condenser_mix.drain
    )
    m.fs.bfpt_to_condmix = Arc(
        source=m.fs.bfpt.outlet,
        destination=m.fs.condenser_mix.bfpt
    )
    m.fs.condmix_to_cond = Arc(
        source=m.fs.condenser_mix.outlet,
        destination=m.fs.condenser.inlet
    )
    m.fs.cond_to_condpump = Arc(
        source=m.fs.condenser.outlet, destination=m.fs.cond_pump.inlet
    )

    # fwh1
    m.fs.condpump_to_fwh1 = Arc(
        source=m.fs.cond_pump.outlet, destination=m.fs.fwh[1].tube_inlet
    )
    m.fs.fwh2_to_fwh1mix = Arc(
        source=m.fs.fwh[2].shell_outlet, destination=m.fs.fwh_mixer[1].drain
    )
    m.fs.fwh1mix_to_fwh1 = Arc(
        source=m.fs.fwh_mixer[1].outlet, destination=m.fs.fwh[1].shell_inlet
    )

    # fwh2
    m.fs.fwh3_to_fwh2mix = Arc(
        source=m.fs.fwh[3].shell_outlet, destination=m.fs.fwh_mixer[2].drain
    )
    m.fs.fwh2mix_to_fwh2 = Arc(
        source=m.fs.fwh_mixer[2].outlet, destination=m.fs.fwh[2].shell_inlet
    )
    m.fs.fwh1_to_fwh2 = Arc(
        source=m.fs.fwh[1].tube_outlet, destination=m.fs.fwh[2].tube_inlet
    )

    # fwh3
    m.fs.fwh4_to_fwh3mix = Arc(
        source=m.fs.fwh[4].shell_outlet, destination=m.fs.fwh_mixer[3].drain
    )
    m.fs.fwh3mix_to_fwh3 = Arc(
        source=m.fs.fwh_mixer[3].outlet, destination=m.fs.fwh[3].shell_inlet
    )
    m.fs.fwh2_to_fwh3 = Arc(
        source=m.fs.fwh[2].tube_outlet, destination=m.fs.fwh[3].tube_inlet
    )

    # fwh4
    m.fs.fwh5_to_fwh4mix = Arc(
        source=m.fs.fwh[5].shell_outlet, destination=m.fs.fwh_mixer[4].drain
    )
    m.fs.fwh4mix_to_fwh4 = Arc(
        source=m.fs.fwh_mixer[4].outlet, destination=m.fs.fwh[4].shell_inlet
    )
    m.fs.fwh3_to_fwh4 = Arc(
        source=m.fs.fwh[3].tube_outlet, destination=m.fs.fwh[4].tube_inlet
    )

    # fwh5
    m.fs.fwh4_to_fwh5 = Arc(
        source=m.fs.fwh[4].tube_outlet, destination=m.fs.fwh[5].tube_inlet
    )

    # Deaerator
    m.fs.fwh5_to_deaerator = Arc(
        source=m.fs.fwh[5].tube_outlet, destination=m.fs.deaerator.feedwater
    )
    m.fs.fwh6_to_deaerator = Arc(
        source=m.fs.fwh[6].shell_outlet, destination=m.fs.deaerator.drain
    )

    # Booster Pump
    m.fs.deaerator_to_booster = Arc(
        source=m.fs.deaerator.outlet, destination=m.fs.booster.inlet
    )

    # fwh6
    m.fs.fwh7_to_fwh6mix = Arc(
        source=m.fs.fwh[7].shell_outlet, destination=m.fs.fwh_mixer[6].drain
    )
    m.fs.fwh6mix_to_fwh6 = Arc(
        source=m.fs.fwh_mixer[6].outlet, destination=m.fs.fwh[6].shell_inlet
    )
    m.fs.booster_to_fwh6 = Arc(
        source=m.fs.booster.outlet, destination=m.fs.fwh[6].tube_inlet
    )

    # fwh7
    m.fs.fwh8_to_fwh7mix = Arc(
        source=m.fs.fwh[8].shell_outlet, destination=m.fs.fwh_mixer[7].drain
    )
    m.fs.fwh7mix_to_fwh7 = Arc(
        source=m.fs.fwh_mixer[7].outlet, destination=m.fs.fwh[7].shell_inlet
    )
    m.fs.fwh6_to_fwh7 = Arc(
        source=m.fs.fwh[6].tube_outlet, destination=m.fs.fwh[7].tube_inlet
    )

    # BFW Pump
    m.fs.fwh7_to_bfp = Arc(
        source=m.fs.fwh[7].tube_outlet, destination=m.fs.bfp.inlet
    )

    # fwh8
    m.fs.fwh9_to_fwh8mix = Arc(
        source=m.fs.fwh[9].shell_outlet, destination=m.fs.fwh_mixer[8].drain
    )
    m.fs.fwh8mix_to_fwh8 = Arc(
        source=m.fs.fwh_mixer[8].outlet, destination=m.fs.fwh[8].shell_inlet
    )
    m.fs.bfp_to_fwh8 = Arc(
        source=m.fs.bfp.outlet, destination=m.fs.fwh[8].tube_inlet
    )

    # fwh9
    m.fs.fwh8_to_fwh9 = Arc(
        source=m.fs.fwh[8].tube_outlet, destination=m.fs.fwh[9].tube_inlet
    )

    # boiler
    m.fs.fwh9_to_boiler = Arc(
        source=m.fs.fwh[9].tube_outlet, destination=m.fs.boiler.inlet
    )


def set_model_input(m):
    # Model inputs / fixed variable or parameter values
    # assumed in this block, unless otherwise stated explicitly,
    # are either assumed or estimated for a total power out of 437 MW
    # These inputs will also fix all necessary inputs to the model
    # i.e. the degrees of freedom = 0

    ###########################################################################
    #  Turbine input                                                          #
    ###########################################################################
    #  Turbine inlet conditions
    m.main_flow = 17854             # Main flow
    m.main_steam_pressure = 31125980
    m.fs.boiler.inlet.flow_mol.fix(m.main_flow)  # mol/s
    m.fs.boiler.outlet.pressure.fix(m.main_steam_pressure)

    # Reheater section pressure drop estimated
    # for a total power out of 437 MW
    m.fs.reheater[1].deltaP.fix(-742845)  # Pa
    m.fs.reheater[2].deltaP.fix(-210952)  # Pa

    # The efficiency and pressure ratios of all turbines were estimated
    # for a total power out of 437 MW
    m.data_turbine_ratioP = {1: 0.388,
                             2: 0.774,
                             3: 0.498,
                             4: 0.609,
                             5: 0.523,
                             6: 0.495,
                             7: 0.514,
                             8: 0.389,
                             9: 0.572,
                             10: 0.476,
                             11: 0.204}
    m.data_turbine_eff = {1: 0.94,
                          2: 0.94,
                          3: 0.94,
                          4: 0.94,
                          5: 0.88,
                          6: 0.88,
                          7: 0.78,
                          8: 0.78,
                          9: 0.78,
                          10: 0.78,
                          11: 0.78}
    for i in m.set_turbine:
        m.fs.turbine[i].ratioP.fix(m.data_turbine_ratioP[i])
        m.fs.turbine[i].efficiency_isentropic.fix(m.data_turbine_eff[i])

    ###########################################################################
    #  Pumps & BFPT                                       #
    ###########################################################################
    m.fs.cond_pump.deltaP.fix(2313881)

    # Unlike the feedwater heaters the steam extraction flow to the deaerator
    # is not constrained by the saturated liquid constraint. Thus, the flow
    # to the deaerator is assumed to be fixed in this model.
    m.fs.turbine_splitter[5].split_fraction[:, "outlet_2"].fix(0.017885)

    # BFW Pump pressure is assumed based on referece report
    m.fs.bfp.outlet.pressure[:].fix(m.main_steam_pressure * 1.1231)  # Pa
    m.fs.booster.deltaP.fix(5715067)

    m.data_pump_eff = 0.8
    for unit in [m.fs.cond_pump, m.fs.booster, m.fs.bfp, m.fs.bfpt]:
        unit.efficiency_isentropic.fix(m.data_pump_eff)

    # Make up stream to condenser
    m.fs.condenser_mix.makeup.flow_mol.value = 1.0E-12  # mol/s
    m.fs.condenser_mix.makeup.pressure.fix(103421.4)  # Pa
    m.fs.condenser_mix.makeup.enth_mol.fix(1131.69204)  # J/mol

    ###########################################################################
    #  FWH section inputs                                        #
    ###########################################################################
    m.data_fwh_area = {1: 250,
                       2: 195,
                       3: 164,
                       4: 208,
                       5: 152,
                       6: 207,
                       7: 202,
                       8: 715,
                       9: 175}

    m.data_fwh_ohtc = {}
    for i in m.set_fwh:
        m.data_fwh_ohtc[i] = 3000

    for i in m.set_fwh:
        m.fs.fwh[i].area.fix(m.data_fwh_area[i])
        m.fs.fwh[i].overall_heat_transfer_coefficient.fix(m.data_fwh_ohtc[i])


def set_scaling_factors(m):
    # scaling factors in the flowsheet

    for i in m.set_fwh:
        b = m.fs.fwh[i]
        iscale.set_scaling_factor(b.area, 1e-2)
        iscale.set_scaling_factor(b.overall_heat_transfer_coefficient, 1e-3)
        iscale.set_scaling_factor(b.shell.heat, 1e-6)
        iscale.set_scaling_factor(b.tube.heat, 1e-6)

    for j in m.set_turbine:
        b = m.fs.turbine[j]
        iscale.set_scaling_factor(b.control_volume.work, 1e-6)

    iscale.set_scaling_factor(m.fs.boiler.control_volume.heat, 1e-6)
    iscale.set_scaling_factor(m.fs.reheater[1].control_volume.heat, 1e-6)
    iscale.set_scaling_factor(m.fs.reheater[2].control_volume.heat, 1e-6)
    iscale.set_scaling_factor(m.fs.condenser.control_volume.heat, 1e-6)
    iscale.set_scaling_factor(m.fs.cond_pump.control_volume.work, 1e-6)
    iscale.set_scaling_factor(m.fs.booster.control_volume.work, 1e-6)
    iscale.set_scaling_factor(m.fs.bfp.control_volume.work, 1e-6)
    iscale.set_scaling_factor(m.fs.bfpt.control_volume.work, 1e-6)


def initialize(m, fileinput=None, outlvl=idaeslog.NOTSET,
               solver=None, optarg={}):

    optarg = {
        "max_iter": 300,
        "halt_on_ampl_error": "yes",
    }
    solver = get_solver(solver, optarg)

    iscale.calculate_scaling_factors(m)

    # # initializing the boiler
    m.fs.boiler.inlet.pressure.fix(32216913)
    m.fs.boiler.inlet.enth_mol.fix(23737)
    m.fs.boiler.initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.boiler.inlet.pressure.unfix()
    m.fs.boiler.inlet.enth_mol.unfix()

    # initialization routine for the turbine train

    # Deactivating constraints that fix enthalpy at FWH outlet
    # This lets us initialize the model using the fixed split_fractions
    # for steam extractions for all the feed water heaters except deaerator
    # These split fractions will be unfixed later and the constraints will
    # be activated
    m.fs.turbine_splitter[1].split_fraction[:, "outlet_2"].fix(0.073444)
    m.fs.turbine_splitter[2].split_fraction[:, "outlet_2"].fix(0.140752)
    m.fs.turbine_splitter[3].split_fraction[:, "outlet_2"].fix(0.032816)
    m.fs.turbine_splitter[4].split_fraction[:, "outlet_2"].fix(0.012425)
    m.fs.turbine_splitter[6].split_fraction[:, "outlet_2"].fix(0.081155)
    m.fs.turbine_splitter[6].split_fraction[:, "outlet_3"].fix(0.091274)
    m.fs.turbine_splitter[7].split_fraction[:, "outlet_2"].fix(0.036058)
    m.fs.turbine_splitter[8].split_fraction[:, "outlet_2"].fix(0.026517)
    m.fs.turbine_splitter[9].split_fraction[:, "outlet_2"].fix(0.029888)
    m.fs.turbine_splitter[10].split_fraction[:, "outlet_2"].fix(0.003007)

    m.fs.constraint_bfp_power.deactivate()
    m.fs.constraint_out_pressure.deactivate()
    for i in m.set_fwh:
        m.fs.fwh[i].fwh_vfrac_constraint.deactivate()

    # solving the turbine, splitter, and reheaters
    propagate_state(m.fs.boiler_to_turb1)
    m.fs.turbine[1].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.turb1_to_t1split)
    m.fs.turbine_splitter[1].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.t1split_to_turb2)
    m.fs.turbine[2].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.turb2_to_t2split)
    m.fs.turbine_splitter[2].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.t2split_to_rh1)
    m.fs.reheater[1].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.rh1_to_turb3)
    m.fs.turbine[3].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.turb3_to_t3split)
    m.fs.turbine_splitter[3].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.t3split_to_turb4)
    m.fs.turbine[4].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.turb4_to_t4split)
    m.fs.turbine_splitter[4].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.t4split_to_rh2)
    m.fs.reheater[2].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.rh2_to_turb5)
    m.fs.turbine[5].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.turb5_to_t5split)
    m.fs.turbine_splitter[5].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.t5split_to_turb6)
    m.fs.turbine[6].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.turb6_to_t6split)
    m.fs.turbine_splitter[6].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.t6split_to_turb7)
    m.fs.turbine[7].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.turb7_to_t7split)
    m.fs.turbine_splitter[7].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.t7split_to_turb8)
    m.fs.turbine[8].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.turb8_to_t8split)
    m.fs.turbine_splitter[8].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.t8split_to_turb9)
    m.fs.turbine[9].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.turb9_to_t9split)
    m.fs.turbine_splitter[9].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.t9split_to_turb10)
    m.fs.turbine[10].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.turb10_to_t10split)
    m.fs.turbine_splitter[10].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.t10split_to_turb11)
    m.fs.turbine[11].initialize(outlvl=outlvl, optarg=solver.options)

    # initialize the boiler feed pump turbine.
    propagate_state(m.fs.t6split_to_bfpt)
    m.fs.bfpt.outlet.pressure.fix(6896)
    m.fs.bfpt.initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.bfpt.outlet.pressure.unfix()

    ###########################################################################
    #  Condenser                                                #
    ###########################################################################
    propagate_state(m.fs.bfpt_to_condmix)
    propagate_state(m.fs.turb11_to_condmix)
    m.fs.condenser_mix.drain.flow_mol.fix(2102)
    m.fs.condenser_mix.drain.pressure.fix(7586)
    m.fs.condenser_mix.drain.enth_mol.fix(3056)
    m.fs.condenser_mix.initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.condenser_mix.drain.unfix()

    propagate_state(m.fs.condmix_to_cond)
    m.fs.condenser.initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.cond_to_condpump)
    m.fs.cond_pump.initialize(outlvl=outlvl, optarg=solver.options)

    ###########################################################################
    #  Low pressure FWH section                                               #
    ###########################################################################

    # fwh1
    propagate_state(m.fs.t10split_to_fwh1mix)
    m.fs.fwh_mixer[1].drain.flow_mol.fix(2072)
    m.fs.fwh_mixer[1].drain.pressure.fix(37187)
    m.fs.fwh_mixer[1].drain.enth_mol.fix(5590)
    m.fs.fwh_mixer[1].initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.fwh_mixer[1].drain.unfix()

    propagate_state(m.fs.fwh1mix_to_fwh1)
    propagate_state(m.fs.condpump_to_fwh1)
    m.fs.fwh[1].initialize(outlvl=outlvl, optarg=solver.options)

    # fwh2
    propagate_state(m.fs.t9split_to_fwh2mix)
    m.fs.fwh_mixer[2].drain.flow_mol.fix(1762)
    m.fs.fwh_mixer[2].drain.pressure.fix(78124)
    m.fs.fwh_mixer[2].drain.enth_mol.fix(7009)
    m.fs.fwh_mixer[2].initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.fwh_mixer[2].drain.unfix()

    propagate_state(m.fs.fwh2mix_to_fwh2)
    propagate_state(m.fs.fwh1_to_fwh2)
    m.fs.fwh[2].initialize(outlvl=outlvl, optarg=solver.options)

    # fwh3
    propagate_state(m.fs.t8split_to_fwh3mix)
    m.fs.fwh_mixer[3].drain.flow_mol.fix(1480)
    m.fs.fwh_mixer[3].drain.pressure.fix(136580)
    m.fs.fwh_mixer[3].drain.enth_mol.fix(8203)
    m.fs.fwh_mixer[3].initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.fwh_mixer[3].drain.unfix()

    propagate_state(m.fs.fwh3mix_to_fwh3)
    propagate_state(m.fs.fwh2_to_fwh3)
    m.fs.fwh[3].initialize(outlvl=outlvl, optarg=solver.options)

    # fwh4
    propagate_state(m.fs.t7split_to_fwh4mix)
    m.fs.fwh_mixer[4].drain.flow_mol.fix(1082)
    m.fs.fwh_mixer[4].drain.pressure.fix(351104)
    m.fs.fwh_mixer[4].drain.enth_mol.fix(10534)
    m.fs.fwh_mixer[4].initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.fwh_mixer[4].drain.unfix()

    propagate_state(m.fs.fwh4mix_to_fwh4)
    propagate_state(m.fs.fwh3_to_fwh4)
    m.fs.fwh[4].initialize(outlvl=outlvl, optarg=solver.options)

    # fwh5
    propagate_state(m.fs.fwh4_to_fwh5)
    propagate_state(m.fs.t6split_to_fwh5)
    m.fs.fwh[5].initialize(outlvl=outlvl, optarg=solver.options)

    ###########################################################################
    #  boiler feed pump and deaerator                                         #
    ###########################################################################
    # Deaerator
    propagate_state(m.fs.fwh5_to_deaerator)
    propagate_state(m.fs.t5split_to_deaerator)
    m.fs.deaerator.drain.flow_mol[:].fix(4277)
    m.fs.deaerator.drain.pressure[:].fix(1379964)
    m.fs.deaerator.drain.enth_mol[:].fix(14898)
    m.fs.deaerator.initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.deaerator.drain.unfix()

    # Booster pump
    propagate_state(m.fs.deaerator_to_booster)
    m.fs.booster.initialize(outlvl=outlvl, optarg=solver.options)

    ###########################################################################
    #  High-pressure feedwater heaters                                        #
    ###########################################################################
    # fwh6
    propagate_state(m.fs.t4split_to_fwh6mix)
    m.fs.fwh_mixer[6].drain.flow_mol.fix(4106)
    m.fs.fwh_mixer[6].drain.pressure.fix(2870602)
    m.fs.fwh_mixer[6].drain.enth_mol.fix(17959)
    m.fs.fwh_mixer[6].initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.fwh_mixer[6].drain.unfix()

    propagate_state(m.fs.fwh6mix_to_fwh6)
    propagate_state(m.fs.booster_to_fwh6)
    m.fs.fwh[6].initialize(outlvl=outlvl, optarg=solver.options)

    # fwh7
    propagate_state(m.fs.t3split_to_fwh7mix)
    m.fs.fwh_mixer[7].drain.flow_mol.fix(3640)
    m.fs.fwh_mixer[7].drain.pressure.fix(4713633)
    m.fs.fwh_mixer[7].drain.enth_mol.fix(20472)
    m.fs.fwh_mixer[7].initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.fwh_mixer[7].drain.unfix()

    propagate_state(m.fs.fwh7mix_to_fwh7)
    propagate_state(m.fs.fwh6_to_fwh7)
    m.fs.fwh[7].initialize(outlvl=outlvl, optarg=solver.options)

    # Boiler feed pump
    propagate_state(m.fs.fwh7_to_bfp)
    m.fs.bfp.initialize(outlvl=outlvl, optarg=solver.options)

    # fwh8
    propagate_state(m.fs.t2split_to_fwh8mix)
    m.fs.fwh_mixer[8].drain.flow_mol.fix(1311)
    m.fs.fwh_mixer[8].drain.pressure.fix(10282256)
    m.fs.fwh_mixer[8].drain.enth_mol.fix(25585)
    m.fs.fwh_mixer[8].initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.fwh_mixer[8].drain.unfix()

    propagate_state(m.fs.fwh8mix_to_fwh8)
    propagate_state(m.fs.bfp_to_fwh8)
    m.fs.fwh[8].initialize(outlvl=outlvl, optarg=solver.options)

    # fwh9
    propagate_state(m.fs.fwh8_to_fwh9)
    propagate_state(m.fs.t1split_to_fwh9)
    m.fs.fwh[9].initialize(outlvl=outlvl, optarg=solver.options)

    #########################################################################
    #  Model Initialization with Square Problem Solve                       #
    #########################################################################
    #  Unfix split fractions and activate vapor fraction constraints
    #  Vaporfrac constraints set condensed steam enthalpy at the condensing
    #  side outlet to be that of a saturated liquid
    # Then solve the square problem again for an initilized model
    for i in m.set_turbine_splitter:
        m.fs.turbine_splitter[i].split_fraction[:, "outlet_2"].unfix()

    # keeping the extraction to deareator to be fixed
    # unfixing the extraction to bfpt
    m.fs.turbine_splitter[5].split_fraction[:, "outlet_2"].fix()
    m.fs.turbine_splitter[6].split_fraction[:, "outlet_3"].unfix()

    m.fs.constraint_bfp_power.activate()
    m.fs.constraint_out_pressure.activate()
    for j in m.set_fwh:
        m.fs.fwh[j].fwh_vfrac_constraint.activate()

    res = solver.solve(m)
    print("Model Initialization = ",
          res.solver.termination_condition)
    print("*******************  USC Model Initialized   ********************")


def add_bounds(m):

    m.flow_max = m.main_flow * 3  # number from Naresh
    # m.salt_flow_max = 1000  # in kg/s

    for unit_k in [
            m.fs.boiler,
            m.fs.reheater[1],
            m.fs.reheater[2],
            m.fs.cond_pump,
            m.fs.bfp,
            m.fs.bfpt
            ]:
        unit_k.inlet.flow_mol[:].setlb(0)  # mol/s
        unit_k.inlet.flow_mol[:].setub(m.flow_max)  # mol/s
        unit_k.outlet.flow_mol[:].setlb(0)  # mol/s
        unit_k.outlet.flow_mol[:].setub(m.flow_max)  # mol/s

    for k in m.set_turbine:
        m.fs.turbine[k].inlet.flow_mol[:].setlb(0)
        m.fs.turbine[k].inlet.flow_mol[:].setub(m.flow_max)
        m.fs.turbine[k].outlet.flow_mol[:].setlb(0)
        m.fs.turbine[k].outlet.flow_mol[:].setub(m.flow_max)

    for k in m.set_fwh_mixer:
        m.fs.fwh_mixer[k].steam.flow_mol[:].setlb(0)
        m.fs.fwh_mixer[k].steam.flow_mol[:].setub(m.flow_max)
        m.fs.fwh_mixer[k].drain.flow_mol[:].setlb(0)
        m.fs.fwh_mixer[k].drain.flow_mol[:].setub(m.flow_max)

    for k in m.set_turbine_splitter:
        m.fs.turbine_splitter[k].split_fraction[0.0, "outlet_1"].setlb(0)
        m.fs.turbine_splitter[k].split_fraction[0.0, "outlet_1"].setub(1)
        m.fs.turbine_splitter[k].split_fraction[0.0, "outlet_2"].setlb(0)
        m.fs.turbine_splitter[k].split_fraction[0.0, "outlet_2"].setub(1)

    for k in m.set_fwh:
        m.fs.fwh[k].shell_inlet.flow_mol[:].setlb(0)
        m.fs.fwh[k].shell_inlet.flow_mol[:].setub(m.flow_max)
        m.fs.fwh[k].tube_inlet.flow_mol[:].setlb(0)
        m.fs.fwh[k].tube_inlet.flow_mol[:].setub(m.flow_max)
        m.fs.fwh[k].shell_outlet.flow_mol[:].setlb(0)
        m.fs.fwh[k].shell_outlet.flow_mol[:].setub(m.flow_max)
        m.fs.fwh[k].tube_outlet.flow_mol[:].setlb(0)
        m.fs.fwh[k].tube_outlet.flow_mol[:].setub(m.flow_max)

    return m


def _add_tags(tag_group, name, state, state_blk, prop_blk):
    if state == "inlet":
        flow, enth, pres, temp, vap = "_Fin", "_Hin", "_Pin", "_Tin", "_xin"
        
    elif state == "outlet":
        flow, enth, pres, temp, vap = "_Fout", "_Hout", "_Pout", "_Tout", "_xout"

    tag_group.add(
        name=name + flow, expr=state_blk.flow_mol[0] * 1e-3, 
        format_string="{:4.3f}", display_units=pyunits.kmol/pyunits.s,
    )
    tag_group.add(
        name=name + enth, expr=state_blk.enth_mol[0] * 1e-3, 
        format_string="{:4.1f}", display_units=pyunits.kJ/pyunits.mol,
    )
    tag_group.add(
        name=name + pres, expr=state_blk.pressure[0] * 1e-6, 
        format_string="{:4.1f}", display_units=pyunits.MPa,
    )
    tag_group.add(
        name=name + temp, expr=prop_blk[0].temperature, 
        format_string="{:4.2f}", display_units=pyunits.K,
    )
    tag_group.add(name=name + vap, expr=prop_blk[0].vapor_frac, format_string="{:4.4f}")


def view_result(outfile, m):
    tags = {}

    tag_group = ModelTagGroup() 

    # Boiler
    tag_group.add(name="power_out", expr=m.fs.plant_power_out[0], format_string="{:4.2f}", display_units=pyunits.MW)

    _add_tags(
        tag_group=tag_group, name="boiler", state="inlet", 
        state_blk=m.fs.boiler.inlet, prop_blk=m.fs.boiler.control_volume.properties_in,
    )
    _add_tags(
        tag_group=tag_group, name="boiler", state="outlet",
        state_blk=m.fs.boiler.outlet, prop_blk=m.fs.boiler.control_volume.properties_out,
    )

    # Reheater 1 & 2
    _add_tags(
        tag_group=tag_group, name="turb3", state="inlet",
        state_blk=m.fs.turbine[3].inlet, prop_blk=m.fs.turbine[3].control_volume.properties_in,
    )
    _add_tags(
        tag_group=tag_group, name="turb5", state="inlet",
        state_blk=m.fs.turbine[5].inlet, prop_blk=m.fs.turbine[5].control_volume.properties_in,
    )

    # Turbine out
    _add_tags(
        tag_group=tag_group, name="turb11", state="outlet",
        state_blk=m.fs.turbine[11].outlet, prop_blk=m.fs.turbine[11].control_volume.properties_out,
    )

    # Condenser
    _add_tags(
        tag_group=tag_group, name="cond", state="outlet",
        state_blk=m.fs.condenser.outlet, prop_blk=m.fs.condenser.control_volume.properties_out,
    )

    # Feed water heaters
    _add_tags(
        tag_group=tag_group, name="fwh9shell", state="inlet",
        state_blk=m.fs.fwh[9].shell_inlet, prop_blk=m.fs.fwh[9].shell.properties_in,
    )
    _add_tags(
        tag_group=tag_group, name="fwh7tube", state="outlet",
        state_blk=m.fs.fwh[7].tube_outlet, prop_blk=m.fs.fwh[7].tube.properties_out,
    )
    _add_tags(
        tag_group=tag_group, name="fwh6shell", state="outlet",
        state_blk=m.fs.fwh[6].shell_outlet, prop_blk=m.fs.fwh[6].shell.properties_out,
    )
    _add_tags(
        tag_group=tag_group, name="fwh5tube", state="outlet",
        state_blk=m.fs.fwh[5].tube_outlet, prop_blk=m.fs.fwh[5].tube.properties_out,
    )
    _add_tags(
        tag_group=tag_group, name="fwh5shell", state="inlet",
        state_blk=m.fs.fwh[5].shell_inlet, prop_blk=m.fs.fwh[5].shell.properties_in,
    )

    # Deareator
    _add_tags(
        tag_group=tag_group, name="da_steam", state="inlet",
        state_blk=m.fs.deaerator.steam, prop_blk=m.fs.deaerator.steam_state,
    )
    _add_tags(
        tag_group=tag_group, name="da", state="outlet",
        state_blk=m.fs.deaerator.outlet, prop_blk=m.fs.deaerator.mixed_state,
    )

    # Feed water heaters mixers
    for i in m.set_fwh_mixer:
        _add_tags(
            tag_group=tag_group, name="fwh" + str(i) + "mix_steam", state="inlet",
            state_blk=m.fs.fwh_mixer[i].steam, prop_blk=m.fs.fwh_mixer[i].steam_state,
        )
        _add_tags(
            tag_group=tag_group, name="fwh" + str(i) + "mix", state="outlet",
            state_blk=m.fs.fwh_mixer[i].outlet, prop_blk=m.fs.fwh_mixer[i].mixed_state,
        )

    # BFP
    tag_group.add(
        name="bfp_power", expr=m.fs.bfp.control_volume.work[0] * 1e-6, 
        format_string="{:4.2f}", display_units=pyunits.MW,
    )
    tag_group.add(
        name="booster_power", expr=m.fs.booster.control_volume.work[0] * 1e-6, 
        format_string="{:4.2f}", display_units=pyunits.MW,
    )
    tag_group.add(
        name="bfpt_power", expr=-m.fs.bfpt.control_volume.work[0] * 1e-6, 
        format_string="{:4.2f}", display_units=pyunits.MW,
    )
    tag_group.add(
        name="cond_power", expr=m.fs.cond_pump.control_volume.work[0] * 1e-6, 
        format_string="{:4.2f}", display_units=pyunits.MW,
    )

    original_svg_file = os.path.join(
        this_file_dir(), "pfd_ultra_supercritical_pc.svg")
    with open(original_svg_file, "r") as f:
        svg_tag(svg=f, tag_group=tag_group, outfile=outfile)


def build_plant_model(m=None):

    # Create a flowsheet, add properties, unit models, and arcs
    m = declare_unit_model(m)

    # Give all the required inputs to the model
    # Ensure that the degrees of freedom = 0 (model is complete)
    set_model_input(m)
    # Assert that the model has no degree of freedom at this point
    assert degrees_of_freedom(m) == 0

    # set scaling factors
    set_scaling_factors(m)

    # adding variable bounds
    add_bounds(m)

    return m


def model_analysis(m, solver):

    #   Solving the flowsheet and check result
    #   At this time one can make chnages to the model for further analysis
    flow_frac_list = [1.0]
    pres_frac_list = [1.0]
    for i in flow_frac_list:
        for j in pres_frac_list:
            m.fs.boiler.inlet.flow_mol.fix(i*17854)  # mol/s
            m.fs.boiler.outlet.pressure.fix(j*31125980)
            solver.solve(m, tee=True, symbolic_solver_labels=True)
            print('Plant Power (MW) =', value(m.fs.plant_power_out[0]))
            print('Plant Heat Duty (MW) =', value(m.fs.plant_heat_duty[0]))

    return m


if __name__ == "__main__":

    optarg = {
        "max_iter": 300,
        "halt_on_ampl_error": "yes"
    }
    solver = get_solver("ipopt", optarg)

    # Build ultra supercriticla power plant model
    m = build_plant_model()

    # Initialize the model (sequencial initialization and custom routines)
    initialize(m)

    # Ensure after the model is initialized, the degrees of freedom = 0
    assert degrees_of_freedom(m) == 0

    # User can import the model from build_plant_model for analysis
    # A sample analysis function is called below
    m_result = model_analysis(m, solver)

    # View results in a process flow diagram
    view_result("pfd_usc_powerplant_result.svg", m_result)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################

import pytest

from pyomo.environ import (
    ConcreteModel,
    value,
    assert_optimal_termination,
)
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog
from dispatches.case_studies.fossil_case.supercritical_plant.supercritical_powerplant import (
    build_scpc_flowsheet,
    fix_dof_and_initialize,
    unfix_dof_for_optimization,
)


@pytest.mark.unit
def test_scpc_without_tes():
    m = build_scpc_flowsheet(include_concrete_tes=False)

    # Check all units are present
    assert not hasattr(m.fs, "tes")
    assert not hasattr(m.fs, "discharge_turbine")
    assert hasattr(m.fs, "boiler")
    assert hasattr(m.fs, "reheater")
    assert hasattr(m.fs, "hp_splitter")
    assert hasattr(m.fs, "bfpt")
    assert hasattr(m.fs, "turbine") and len(m.fs.turbine) == 9
    assert hasattr(m.fs, "t_splitter") and len(m.fs.t_splitter) == 8

    fix_dof_and_initialize(m, outlvl=idaeslog.WARNING)

    res = get_solver().solve(m)
    assert_optimal_termination(res)

    unfix_dof_for_optimization(m)

    assert (m.fs.net_power_output[0].value / 1e6) == pytest.approx(692, abs=1)


@pytest.mark.unit
def test_scpc_with_tes():
    m = ConcreteModel()
    m = build_scpc_flowsheet(m=m, include_concrete_tes=True)

    # Check all units are present
    assert hasattr(m.fs, "tes")
    assert hasattr(m.fs, "discharge_turbine")

    fix_dof_and_initialize(m)

    res = get_solver().solve(m)
    assert_optimal_termination(res)

    unfix_dof_for_optimization(m)

    assert (m.fs.net_power_output[0].value / 1e6) == pytest.approx(625, abs=1)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
import numpy as np
import pandas as pd
from pathlib import Path

def read_prescient_outputs(output_dir, source_dir, gen_name=None):
    """
    Read the bus information, bus LMPs, and renewables output detail CSVs from Prescient simulations

    Args:
        output_dir: Prescient simulation output directory
        source_dir: "SourceData" folder of the RTS-GMLC
    """
    source_dir = Path(source_dir)
    output_dir = Path(output_dir)

    summary = pd.read_csv(output_dir / "hourly_summary.csv")
    summary['Datetime'] = pd.to_datetime(summary['Date'].astype('str') + " " + summary['Hour'].astype('str') + ":00", format='%Y-%m-%d %H:%M')
    summary = summary.set_index(pd.DatetimeIndex(summary['Datetime']))
    summary = summary.drop(['Date', 'Hour', 'Datetime'], axis=1)
    
    bus_detail_df = pd.read_csv(output_dir / "bus_detail.csv")
    bus_detail_df['LMP'] = bus_detail_df['LMP'].astype('float64')
    bus_detail_df['LMP DA'] = bus_detail_df['LMP DA'].astype('float64')
    bus_detail_df['Datetime'] = pd.to_datetime(bus_detail_df['Date'].astype('str') + " " + bus_detail_df['Hour'].astype('str') + ":" + bus_detail_df['Minute'].astype('str'), format='%Y-%m-%d %H:%M')
    bus_detail_df = bus_detail_df.set_index(pd.DatetimeIndex(bus_detail_df['Datetime']))
    bus_detail_df = bus_detail_df.drop(['Date', 'Hour', 'Minute', 'Datetime'], axis=1)

    summary = pd.merge(summary, bus_detail_df, how='outer', on=['Datetime'])

    renewables_df = pd.read_csv(output_dir / "renewables_detail.csv")
    thermal_df = pd.read_csv(output_dir / 'thermal_detail.csv')

    load_renewables = gen_name == None or gen_name in renewables_df.Generator.unique()
    load_thermal = gen_name == None or gen_name in thermal_df.Generator.unique()
    
    if load_renewables:
        renewables_df['Datetime'] = pd.to_datetime(renewables_df['Date'].astype('str') + " " + renewables_df['Hour'].astype('str') + ":" + renewables_df['Minute'].astype('str'), format='%Y-%m-%d %H:%M')
        renewables_df = renewables_df.set_index(pd.DatetimeIndex(renewables_df['Datetime']))
        renewables_df = renewables_df.drop(['Date', 'Hour', 'Minute', 'Datetime'], axis=1)
        gen_df = renewables_df

    if load_thermal:
        thermal_df['Datetime'] = pd.to_datetime(thermal_df['Date'].astype('str') + " " + thermal_df['Hour'].astype('str') + ":" + thermal_df['Minute'].astype('str'), format='%Y-%m-%d %H:%M')
        thermal_df = thermal_df.set_index(pd.DatetimeIndex(thermal_df['Datetime']))
        thermal_df = thermal_df.drop(['Date', 'Hour', 'Minute', 'Datetime'], axis=1)
        gen_df = thermal_df

    if load_thermal and load_renewables:
        gen_df = pd.merge(renewables_df, thermal_df, how='outer', on=['Datetime', 'Generator'])

    return summary, gen_df


def read_rts_gmlc_wind_inputs(source_dir, gen_name=None):
    """
    Read capacity factors for day ahead and real time wind forecasts. 
    The forecasts are provided as 12 periods per hour, or 288 per day, and are resampled to hourly data.
    The time index is shifted by one day relative to the other time series data, so np.roll is required at the end.

    Args:
        source_dir: "SourceData" folder of the RTS-GMLC
        gen_name: optional wind generator name, if not provided then all wind generators returned
    """
    source_dir = Path(source_dir)
    gen_df = pd.read_csv(source_dir / "gen.csv")
    if not gen_name:
        wind_gens = [i for i in gen_df['GEN UID'] if "WIND" in i]
    else:
        wind_gens = [gen_name]

    wind_rt_df = pd.read_csv(source_dir.parent / "timeseries_data_files" / "WIND" / "REAL_TIME_wind.csv")
    wind_da_df = pd.read_csv(source_dir.parent / "timeseries_data_files" / "WIND" / "DAY_AHEAD_wind.csv")

    start_df = wind_rt_df.head(1)
    start_year = start_df.Year.values[0]
    start_mon = start_df.Month.values[0]
    start_day = start_df.Day.values[0]

    start_date = pd.Timestamp(f"{start_year}-{start_mon:02d}-{start_day:02d} 00:00:00")
    ix = pd.date_range(start=start_date, 
                        end=start_date
                        + pd.offsets.DateOffset(days=366)
                        - pd.offsets.DateOffset(hours=1),
                        freq='1H')
    wind_df = pd.DataFrame(index=ix)
    for k in wind_gens:
        rt_wind = wind_rt_df[k].values
        rt_wind = np.reshape(rt_wind, (8784, 12))
        rt_wind = rt_wind.mean(1)
        rt_wind = np.roll(rt_wind, 1)
        da_wind = np.roll(wind_da_df[k].values, 1)

        wind_pmax = gen_df[gen_df['GEN UID'] == k]['PMax MW'].values[0]
        wind_df[k+"-RTCF"] = rt_wind / wind_pmax
        wind_df[k+"-DACF"] = da_wind / wind_pmax
    return wind_df


def prescient_outputs_for_gen(output_dir, source_dir, gen_name):
    """
    Get timeseries RT and DA Outputs and Capacity factors, Curtailment, Unit Market Revenue, Unit Uplift Payment for the given generator,
    and also the Demand, Shortfall, Overgeneration, and LMPs for the bus of the generator

    Args:
        output_dir: Prescient simulation output directory
        source_dir: "SourceData" folder of the RTS-GMLC
        gen_name: optional wind generator name, if not provided then all wind generators returned
    """
    source_dir = Path(source_dir)
    output_dir = Path(output_dir)
    summary, gen_df = read_prescient_outputs(output_dir, source_dir, gen_name)
    if "WIND" in gen_name or "PV" in gen_name:
        # double loop may have set the wind or pv plant as a thermal generator, so then these columns are not meaningful
        summary = summary.drop(['RenewablesUsed', 'RenewablesCurtailment'], axis=1)

    bus_names = pd.read_csv(source_dir / "bus.csv")
    bus_dict = {k: v for k, v in zip(bus_names['Bus ID'].values, bus_names['Bus Name'].values)}
    bus_name = bus_dict[int(gen_name.split('_')[0])]

    summary = summary[summary.Bus == bus_name]
    gen_df = gen_df[gen_df.Generator == gen_name]
    df = pd.concat([summary, gen_df], axis=1)

    if 'WIND' in gen_name:
        wind_forecast_df = read_rts_gmlc_wind_inputs(source_dir, gen_name)
        wind_forecast_df = wind_forecast_df[wind_forecast_df.index.isin(df.index)]
        df = pd.concat([df, wind_forecast_df], axis=1)
    return df


def prescient_double_loop_outputs_for_gen(output_dir):
    """
    Get timeseries double-loop simulation outputs
    """
    output_dir = Path(output_dir)
    tracker_df = pd.read_csv(output_dir / "tracker_detail.csv")
    tracker_df['Datetime'] = pd.to_datetime(tracker_df['Date'].astype('str') + " " + tracker_df['Hour'].astype('str') + ":00", format='%Y-%m-%d %H:%M')
    tracker_df = tracker_df.set_index(pd.DatetimeIndex(tracker_df['Datetime']))
    tracker_df = tracker_df.drop(['Date', 'Hour', 'Datetime'], axis=1)

    tracker_model_df = pd.read_csv(output_dir / "tracking_model_detail.csv")
    tracker_model_df['Datetime'] = pd.to_datetime(tracker_model_df['Date'].astype('str') + " " + tracker_model_df['Hour'].astype('str') + ":00", format='%Y-%m-%d %H:%M')
    tracker_model_df = tracker_model_df.set_index(pd.DatetimeIndex(tracker_model_df['Datetime']))
    gen_name = tracker_model_df['Generator'].values[0]
    tracker_model_df = tracker_model_df.drop(['Date', 'Hour', 'Datetime', 'Generator'], axis=1)

    tracker_df = pd.merge(tracker_df, tracker_model_df, how='left', on=['Datetime', 'Horizon [hr]'])
    tracker_df.loc[:, 'Model'] = 'Tracker'

    bidder_df = pd.read_csv(output_dir / "bidder_detail.csv")
    da_bidder_df = bidder_df[bidder_df['Market'] == 'Day-ahead']
    da_bidder_df = da_bidder_df.rename(columns={'Hour': 'Horizon [hr]', 'Market': 'Model', 'Power 0 [MW]': 'DA Power 0 [MW]', 'Cost 0 [$]': 'DA Cost 0 [$]', 'Power 1 [MW]': 'DA Power 1 [MW]',
       'Cost 1 [$]': 'DA Cost 1 [$]', 'Power 2 [MW]': 'DA Power 2 [MW]', 'Cost 2 [$]': 'DA Cost 2 [$]'})
    da_bidder_df.loc[:, 'Model'] = 'DA Bidder'
    da_bidder_df['Datetime'] = pd.to_datetime(da_bidder_df['Date'].astype('str') + " 00:00:00", format='%Y-%m-%d %H:%M:%S')
    da_bidder_df = da_bidder_df.set_index(pd.DatetimeIndex(da_bidder_df['Datetime']))
    da_bidder_df = da_bidder_df.drop(['Date', 'Datetime', 'Generator'], axis=1)

    rt_bidder_df = bidder_df[bidder_df['Market'] == 'Real-time']
    rt_bidder_df = rt_bidder_df.rename(columns={'Market': 'Model', 'Power 0 [MW]': 'RT Power 0 [MW]', 'Cost 0 [$]': 'RT Cost 0 [$]', 'Power 1 [MW]': 'RT Power 1 [MW]',
       'Cost 1 [$]': 'RT Cost 1 [$]', 'Power 2 [MW]': 'RT Power 2 [MW]', 'Cost 2 [$]': 'RT Cost 2 [$]'})
    rt_bidder_df.loc[:, 'Model'] = 'RT Bidder'
    rt_bidder_df['Hour'] = np.tile(np.repeat(range(24), 4), 366)[0:len(rt_bidder_df)]
    rt_bidder_df['Horizon [hr]'] = np.tile(range(4), 8784)[0:len(rt_bidder_df)]
    rt_bidder_df['Datetime'] = pd.to_datetime(rt_bidder_df['Date'].astype('str') + " " + rt_bidder_df['Hour'].astype('str') + ":00", format='%Y-%m-%d %H:%M')
    rt_bidder_df = rt_bidder_df.set_index(pd.DatetimeIndex(rt_bidder_df['Datetime']))
    rt_bidder_df = rt_bidder_df.drop(['Date', 'Hour', 'Datetime', 'Generator'], axis=1)

    bidder_df = pd.merge(da_bidder_df, rt_bidder_df, how='outer', on=['Datetime', 'Horizon [hr]', 'Model'])
    df = pd.merge(bidder_df, tracker_df, how='outer', on=['Datetime', 'Horizon [hr]', 'Model'])
    return df, gen_name


def double_loop_outputs_for_gen(double_loop_dir, source_dir):
    source_dir = Path(source_dir)
    double_loop_dir = Path(double_loop_dir)
    dl_df, gen_name = prescient_double_loop_outputs_for_gen(double_loop_dir)
    res_df = prescient_outputs_for_gen(double_loop_dir, source_dir, gen_name)
    res_df = res_df[res_df.index.isin(dl_df.index.unique())]
    res_df['Datetime'] = res_df.index
    res_df.loc[:, 'Model'] = "Prescient"

    df = pd.merge(dl_df, res_df, how='outer', on=['Datetime', 'Model']).set_index('Datetime')
    df.Model = pd.Categorical(df.Model, categories=['DA Bidder', 'RT Bidder', 'Tracker', "Prescient"])
    df = df.sort_values(by = ['Datetime', 'Model'], na_position = 'last')
    return df.dropna(axis=1, how='all')



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
import pytest
from pathlib import Path
import pandas as pd
import numpy as np
import pyomo.environ as pyo
from pyomo.common.fileutils import this_file_dir
from pyomo.common.unittest import assertStructuredAlmostEqual

from idaes.apps.grid_integration.tracker import Tracker
from idaes.apps.grid_integration.bidder import SelfScheduler, Bidder
from idaes.apps.grid_integration.forecaster import Backcaster
from idaes.apps.grid_integration.model_data import (
    RenewableGeneratorModelData,
    ThermalGeneratorModelData,
)
from dispatches.case_studies.renewables_case.wind_battery_double_loop import MultiPeriodWindBattery


@pytest.fixture
def wind_thermal_dispatch_data():
    re_case_dir = Path(this_file_dir()).parent
    df = pd.read_csv(re_case_dir / "data" / "Wind_Thermal_Dispatch.csv")
    df["DateTime"] = df['Unnamed: 0']
    df.drop('Unnamed: 0', inplace=True, axis=1)
    df.index = pd.to_datetime(df["DateTime"])
    return df


@pytest.mark.component
def test_track_market_dispatch(wind_thermal_dispatch_data):

    tracking_horizon = 4
    pmin = 0
    pmax = 200

    solver = pyo.SolverFactory("ipopt")

    generator_params = {
        "gen_name": "309_WIND_1",
        "bus": "Carter",
        "p_min": pmin,
        "p_max": pmax,
        "p_cost": 0,
        "fixed_commitment": None,
    }
    model_data = RenewableGeneratorModelData(**generator_params)

    mp_wind_battery = MultiPeriodWindBattery(
        model_data=model_data,
        wind_capacity_factors=wind_thermal_dispatch_data["309_WIND_1-RTCF"].values,
        wind_pmax_mw=pmax,
        battery_pmax_mw=25,
        battery_energy_capacity_mwh=100,
    )

    n_tracking_hour = 1

    # create a `Tracker` using`mp_wind_battery`
    tracker_object = Tracker(
        tracking_model_object=mp_wind_battery,
        tracking_horizon=tracking_horizon,
        n_tracking_hour=n_tracking_hour,
        solver=solver,
    )

    # example market dispatch signal for 4 hours
    market_dispatch = [0, 1.5, 15.0, 24.5]

    # find a solution that tracks the dispatch signal
    tracker_object.track_market_dispatch(
        market_dispatch=market_dispatch, date="2020-01-02", hour="00:00"
    )

    blks = tracker_object.model.fs.windBattery.get_active_process_blocks()
    assert len(blks) == tracking_horizon

    wind_power = [blks[i].fs.windpower.electricity[0].value for i in range(4)]
    expected_wind_power = [1123.8, 1573.4, 20510.2, 25938.4]

    for power, expected in zip(wind_power, expected_wind_power):
        assert (
            pytest.approx(power, rel=1e-3) == expected
        )

    produced_power = [pyo.value(tracker_object.power_output[t])
                      for t in range(tracking_horizon)]

    for power, expected in zip(produced_power, market_dispatch):
        assert (
            pytest.approx(power, abs=1e-3) == expected
        )

    expected_battery_energy = [expected_wind_power[i] -
                               market_dispatch[i] * 1e3 for i in range(tracking_horizon)]
    battery_power = [blks[i].fs.battery.elec_in[0].value for i in range(4)]

    for power, expected in zip(battery_power, expected_battery_energy):
        assert (
            pytest.approx(power, rel=1e-3) == expected
        )


def test_compute_bids_self_schedule(wind_thermal_dispatch_data):
    day_ahead_horizon = 48
    real_time_horizon = 4
    n_scenario = 1
    pmin = 0
    pmax = 200
    bus_name = "Carter"

    historical_da_prices = wind_thermal_dispatch_data["309_DALMP"].values[0:48].tolist(
    )
    historical_rt_prices = wind_thermal_dispatch_data["309_RTLMP"].values[0:48].tolist(
    )

    backcaster = Backcaster({bus_name: historical_da_prices}, {
                            bus_name: historical_rt_prices})

    generator_params = {
        "gen_name": "309_WIND_1",
        "bus": bus_name,
        "p_min": pmin,
        "p_max": pmax,
        "p_cost": 0,
        "fixed_commitment": None,
    }
    model_data = RenewableGeneratorModelData(**generator_params)

    solver = pyo.SolverFactory("cbc")

    mp_wind_battery_bid = MultiPeriodWindBattery(
        model_data=model_data,
        wind_capacity_factors=wind_thermal_dispatch_data["309_WIND_1-RTCF"].values,
        wind_pmax_mw=pmax,
        battery_pmax_mw=25,
        battery_energy_capacity_mwh=100,
    )

    bidder_object = SelfScheduler(
        bidding_model_object=mp_wind_battery_bid,
        day_ahead_horizon=day_ahead_horizon,
        real_time_horizon=real_time_horizon,
        n_scenario=n_scenario,
        solver=solver,
        forecaster=backcaster,
    )

    date = "2020-01-02"
    bid_energies = bidder_object.compute_day_ahead_bids(date=date)
    bid_energies = [i['309_WIND_1']['p_max'] for i in bid_energies.values()]

    blks = bidder_object.day_ahead_model.fs[0].windBattery.get_active_process_blocks(
    )
    assert len(blks) == day_ahead_horizon
    assert len(bidder_object.day_ahead_model.fs.index_set()) == n_scenario

    known_solution = [
        0.0, 1.5734, 0.0, 0.0, 10.0865, 30.7449, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 11.9699, 1.3711, 4.7876, 20.5439, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 86.0643, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 35.7721
    ]

    assertStructuredAlmostEqual(bid_energies, known_solution, reltol=1e-2)


def test_compute_bids_thermal_gen(wind_thermal_dispatch_data):
    day_ahead_horizon = 48
    real_time_horizon = 4
    n_scenario = 1
    pmin = 0
    wind_pmax = 200
    battery_pmax = 25
    bus_name = "Carter"

    historical_da_prices = wind_thermal_dispatch_data["309_DALMP"].values[0:48].tolist(
    )
    historical_rt_prices = wind_thermal_dispatch_data["309_RTLMP"].values[0:48].tolist(
    )

    backcaster = Backcaster({bus_name: historical_da_prices}, {
                            bus_name: historical_rt_prices})

    thermal_generator_params = {
        "gen_name": "309_WIND_1",
        "bus": bus_name,
        "p_min": pmin,
        "p_max": wind_pmax,
        "min_down_time": 0,
        "min_up_time": 0,
        "ramp_up_60min": wind_pmax + battery_pmax,
        "ramp_down_60min": wind_pmax + battery_pmax,
        "shutdown_capacity": wind_pmax + battery_pmax,
        "startup_capacity": 0,
        "initial_status": 1,
        "initial_p_output": 0,
        "production_cost_bid_pairs": [(pmin, 0), (wind_pmax, 0)],
        "startup_cost_pairs": [(0, 0)],
        "fixed_commitment": None,
    }
    model_data = ThermalGeneratorModelData(**thermal_generator_params)

    solver = pyo.SolverFactory("cbc")

    mp_wind_battery_bid = MultiPeriodWindBattery(
        model_data=model_data,
        wind_capacity_factors=wind_thermal_dispatch_data["309_WIND_1-RTCF"].values,
        wind_pmax_mw=wind_pmax,
        battery_pmax_mw=battery_pmax,
        battery_energy_capacity_mwh=battery_pmax * 4,
    )

    bidder_object = Bidder(
        bidding_model_object=mp_wind_battery_bid,
        day_ahead_horizon=day_ahead_horizon,
        real_time_horizon=real_time_horizon,
        n_scenario=n_scenario,
        solver=solver,
        forecaster=backcaster,
    )

    date = "2020-01-02"
    bids = bidder_object.compute_day_ahead_bids(date=date)
    bids = [i['309_WIND_1']['p_cost'] for i in bids.values()]
    bid_prices = [bid[-1][1] for bid in bids]
    print(bid_prices)

    blks = bidder_object.day_ahead_model.fs[0].windBattery.get_active_process_blocks(
    )
    assert len(blks) == day_ahead_horizon
    assert len(bidder_object.day_ahead_model.fs.index_set()) == n_scenario

    known_solution = [
        0.0, 6188.0, 0.0, 0.0, 5270.0, 6132.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 
        0.0, 0.0, 0.0, 7502.0, 7224.0, 6750.000000000001, 5358.0, 0.0, 0.0, 0.0, 0.0, 
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 
        3772.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 3938.0]

    assertStructuredAlmostEqual(bid_prices, known_solution, reltol=1e-2)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
import sys
from pandas import DataFrame
from collections import OrderedDict
import textwrap
# Import Pyomo libraries
from pyomo.environ import Var, Param, NonNegativeReals, units as pyunits, value
from pyomo.network import Port
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        UnitModelBlockData)
from idaes.core.solvers import get_solver
from idaes.core.util.tables import stream_table_dataframe_to_string
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_activated_constraints,
                                              number_activated_blocks)
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("BatteryStorage", doc="Simple battery model")
class BatteryStorageData(UnitModelBlockData):
    """
    Wind plant using turbine powercurve and resource data.
    Unit model to convert wind resource into electricity.
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Battery does not support dynamic models, thus this must be False."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag",
        doc="""Battery does not have defined volume, thus this must be False."""))

    def build(self):
        """Building model
        This model does not use the flowsheet's time domain. Instead, it only models a single timestep, with initial
        conditions provided by `initial_state_of_charge` and `initial_energy_throughput`. The model calculates change
        in stored energy across a single time step using the power flow variables, `power_in` and `power_out`, and
        the `dr_hr` parameter.
        Args:
            None
        Returns:
            None
        """
        super().build()

        # Design variables and parameters
        self.nameplate_power = Var(within=NonNegativeReals,
                                   initialize=0.0,
                                   bounds=(0, 1e8),
                                   doc="Nameplate power of battery energy storage",
                                   units=pyunits.kW)

        self.nameplate_energy = Var(within=NonNegativeReals,
                                    initialize=0.0,
                                    bounds=(0, 1e9),
                                    doc="Nameplate energy of battery energy storage",
                                    units=pyunits.kWh)

        self.charging_eta = Param(within=NonNegativeReals,
                                  mutable=True,
                                  initialize=0.95,
                                  doc="Charging efficiency, (0, 1]")

        self.discharging_eta = Param(within=NonNegativeReals,
                                     mutable=True,
                                     initialize=0.95,
                                     doc="Discharging efficiency, (0, 1]")

        self.degradation_rate = Param(within=NonNegativeReals,
                                      mutable=True,
                                      initialize=1/10000,
                                      doc="Degradation rate, [0, 1e-3]",
                                      units=pyunits.hr/pyunits.hr)

        # Initial conditions
        self.initial_state_of_charge = Var(within=NonNegativeReals,
                                           initialize=0.0,
                                           doc="State of charge at t - 1, [0, self.nameplate_energy]",
                                           units=pyunits.kWh)

        self.initial_energy_throughput = Var(within=NonNegativeReals,
                                             initialize=0.0,
                                             doc="Cumulative energy throughput at t - 1",
                                             units=pyunits.kWh)

        # Power flows and energy storage
        self.dt = Param(within=NonNegativeReals,
                        initialize=1,
                        doc="Time step for converting between electricity power flows and stored energy",
                        units=pyunits.hr)

        self.elec_in = Var(self.flowsheet().config.time,
                           within=NonNegativeReals,
                           initialize=0.0,
                           doc="Energy in",
                           units=pyunits.kW)

        self.elec_out = Var(self.flowsheet().config.time,
                            within=NonNegativeReals,
                            initialize=0.0,
                            doc="Energy out",
                            units=pyunits.kW)

        self.state_of_charge = Var(self.flowsheet().config.time,
                                   within=NonNegativeReals,
                                   initialize=0.0,
                                   doc="State of charge (energy), [0, self.nameplate_energy]",
                                   units=pyunits.kWh)

        self.energy_throughput = Var(self.flowsheet().config.time,
                                     within=NonNegativeReals,
                                     initialize=0.0,
                                     doc="Cumulative energy throughput",
                                     units=pyunits.kWh)

        # Ports
        self.power_in = Port(noruleinit=True, doc="A port for electricity inflow")
        self.power_in.add(self.elec_in, "electricity")

        self.power_out = Port(noruleinit=True, doc="A port for electricity outflow")
        self.power_out.add(self.elec_out, "electricity")

        @self.Constraint(self.flowsheet().config.time)
        def state_evolution(b, t):
            return b.state_of_charge[t] == b.initial_state_of_charge + (
                    b.charging_eta * b.dt * b.elec_in[t]
                    - b.dt / b.discharging_eta * b.elec_out[t])

        @self.Constraint(self.flowsheet().config.time)
        def accumulate_energy_throughput(b, t):
            return b.energy_throughput[t] == b.initial_energy_throughput + b.dt * (b.elec_in[t] + b.elec_out[t]) / 2

        @self.Constraint(self.flowsheet().config.time)
        def state_of_charge_bounds(b, t):
            return b.state_of_charge[t] <= b.nameplate_energy - b.degradation_rate * b.energy_throughput[t]

        @self.Constraint(self.flowsheet().config.time)
        def power_bound_in(b, t):
            return b.elec_in[t] <= b.nameplate_power

        @self.Constraint(self.flowsheet().config.time)
        def power_bound_out(b, t):
            return b.elec_out[t] <= b.nameplate_power

    def initialize_build(self, outlvl=idaeslog.NOTSET, solver=None, optarg=None):
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl,
                                            tag="properties")
        opt = get_solver(solver=solver, options=optarg)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info("Battery initialization status {}.".
                      format(idaeslog.condition(res)))

    def report(self, time_point=0, dof=False, ostream=None, prefix=""):
        time_point = float(time_point)

        if ostream is None:
            ostream = sys.stdout

        # Get DoF and model stats
        if dof:
            dof_stat = degrees_of_freedom(self)
            nv = number_variables(self)
            nc = number_activated_constraints(self)
            nb = number_activated_blocks(self)

        # Get stream table
        stream_attributes = OrderedDict()
        stream_attributes["Inlet"] = {'electricity': value(self.elec_in[time_point])}
        stream_attributes["Outlet"] = {'electricity': value(self.elec_out[time_point])}
        stream_attributes["kWh"] = {}
        stream_attributes["kWh"]['initial_state_of_charge'] = value(self.initial_state_of_charge)
        stream_attributes["kWh"]['initial_energy_throughput'] = value(self.initial_energy_throughput)
        stream_attributes["kWh"]['state_of_charge'] = value(self.state_of_charge[time_point])
        stream_attributes["kWh"]['energy_throughput'] = value(self.energy_throughput[time_point])

        stream_table = DataFrame.from_dict(stream_attributes, orient="columns")

        if hasattr(self, "is_flowsheet") and self.is_flowsheet:
            model_type = "Flowsheet"
        else:
            model_type = "Unit"

        max_str_length = 84
        tab = " " * 4
        ostream.write("\n" + "=" * max_str_length + "\n")

        lead_str = f"{prefix}{model_type} : {self.name}"
        trail_str = f"Time: {time_point}"
        mid_str = " " * (max_str_length - len(lead_str) - len(trail_str))
        ostream.write(lead_str + mid_str + trail_str)

        if dof:
            ostream.write("\n" + "=" * max_str_length + "\n")
            ostream.write(f"{prefix}{tab}Local Degrees of Freedom: {dof_stat}")
            ostream.write('\n')
            ostream.write(f"{prefix}{tab}Total Variables: {nv}{tab}"
                          f"Activated Constraints: {nc}{tab}"
                          f"Activated Blocks: {nb}")

        if stream_table is not None:
            ostream.write("\n" + "-" * max_str_length + "\n")
            ostream.write(f"{prefix}{tab}Stream Table")
            ostream.write('\n')
            ostream.write(
                textwrap.indent(
                    stream_table_dataframe_to_string(stream_table),
                    prefix + tab))
        ostream.write("\n" + "=" * max_str_length + "\n")
